#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Classes for SLA Pages / Master pages.
"""

# Scribus pages are  arranged in a vast scratch space,  where 1 unit ==
# 1/72 inch.  Positive x is to  the right, and positive  y is downward.
# The positions, width and height in  the elements are in scratch space
# coordinates  with the  exception  of POCOOR  and COCOOR  coordinates,
# which are in the rotated and translated space of a PAGEOBJECT.

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from typing import Optional, NoReturn, Literal

from enum import Enum, IntEnum

import lxml
import lxml.etree as ET

import pyscribus.logs as logs
import pyscribus.common.xml as xmlc
import pyscribus.common.math as mathc
import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

PageOrientation = Literal["portrait", "landscape"]

# FIXME Replace with Size(StrEnum) introduced in Python 3.11at some point
# in which the values are separated by commas

xml_size = {
    "A4": 'A4',
    "LETTER": 'letter',
    "CUSTOM": 'custom'
}

# Classes ===============================================================#


class Orientation(IntEnum):
    """
    Enum for DOC/@ORIENTATION values.
    """

    PORTRAIT = 0
    LANDSCAPE = 1


class PageAbstract(xmlc.PyScribusElement):
    """
    Abstract class for pages and master pages.

    Use Page or MasterPage classes instead.

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table in Page, MasterPage)

    .. seealso:: :class:`MasterPage`, :class:`Page`
    """

    orientation_xml = {"portrait": "0", "landscape": "1"}
    autoguides_origin_xml = {"page": "0", "margins": "1", "selection": "2"}

    effect_source_xml = {"internal": "0", "external": "1"}
    effect_mobile_line_xml = {"horizontal": "0", "vertical": 1}

    effect_type_xml = {
        "none": "0",
        "masks": "1",
        "box": "2",
        "dissolve": "3",
        "glitter": "4",
        "break": "5",
        "delete": "6"
    }

    # NOTE These directions are in the correct order, but Scribus only allows
    # the user to select "left-to-right" in its GUI.
    effect_direction_xml = {
        "left-to-right": "0",
        "top-to-bottom": "1",
        "bottom-to-top": "2",
        "right-to-left": "3",
        "tl-to-br": "4",
    }

    def __init__(self, **kwargs):
        xmlc.PyScribusElement.__init__(self)

        self.box = dimensions.DimBox()

        self.borders = {
            "top": dimensions.Dim(40),
            "left": dimensions.Dim(40),
            "right": dimensions.Dim(40),
            "bottom": dimensions.Dim(40),
        }

        self.name = ""
        self.master_name = ""

        self.orientation = "portrait"

        self.paper_size = ""

        self.number = -1

        #--- FIXME Tester pour savoir comment constituer les listes ------

        self.guides = {
            "horizontal": [],
            "vertical": []
        }

        # ----------------------------------------------------------------

        # @LEFT - For multipage spreads, which page in the spread is the left most
        self.is_leftest = False

        # FIXME Not documented -------------------------------------------

        # PRESET="0"

        #--- Auto guides -------------------------------------------------

        self.auto_guides = {
            "lines": {
                # @AGhorizontalAutoCount
                "count": 0,
                # @AGhorizontalAutoGap
                "gap": dimensions.Dim(0),
                # @AGhorizontalAutoRefer
                "origin": "page"
            },
            "columns": {
                # @AGverticalAutoCount
                "count": 0,
                # @AGverticalAutoGap
                "gap": dimensions.Dim(0),
                # @AGverticalAutoRefer
                "origin": "page",
            },
            # @AGSelection
            # In XML, this is a string of floats X, Y, Width, Height separated
            # by spaces.
            "selection": {
                "posx": dimensions.Dim(0),
                "posy": dimensions.Dim(0),
                "width": dimensions.Dim(0),
                "height": dimensions.Dim(0),
            }
        }

        #--- PDF effects -------------------------------------------------

        self.effect = {
            # @pageEffectDuration
            "duration": dimensions.Dim(1, unit="sec"),
            # @pageViewDuration
            "view-duration": dimensions.Dim(1, unit="sec"),
            # @effectType
            "type": "none",
            # @Dm
            "mobile-lines": "horizontal",
            # @M
            "source": "internal",
            # @Di
            "direction": "left-to-right"
        }

        #--- FIXME Not documented ----------------------------------------

        # Same as DOCUMENT/PDF/Effekte (PDF display effects) attributes ?

        #-----------------------------------------------------------------

        if kwargs:
            self._quick_setup(kwargs)

    def _quick_setup(self, settings):
        """
        Method for defining (master)page settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            xmlc.PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                # (Master) page box -------------------------------------------

                if setting_name == "posx":
                    # self.box.coords["top-left"][0].value = float(setting_value)
                    self.box.setx("top-left", float(setting_value))

                if setting_name == "posy":
                    # self.box.coords["top-left"][1].value = float(setting_value)
                    self.box.sety("top-left", float(setting_value))

                if setting_name == "width":
                    self.box.dims["width"].value = float(setting_value)

                if setting_name == "height":
                    self.box.dims["height"].value = float(setting_value)

                # Borders -----------------------------------------------------

                if setting_name in [
                        "rightborder", "leftborder",
                        "topborder", "bottomborder"]:
                    side = setting_name.split("border")[0]
                    self.borders[side].value = float(setting_value)

                if setting_name == "borders":
                    # NOTE We solve borders settings like CSS margin property
                    # If only one value : same border for all sides
                    # If two values: vertical and horizontal borders
                    # If three values: top horizontal bottom
                    # If four values: top right bottom left

                    if isinstance(setting_value, list):
                        setting_len = len(setting_value)

                        if setting_len == 1:

                            for side in ["top", "right", "bottom", "left"]:
                                self.borders[side].value = float(
                                    setting_value[0]
                                )

                        if setting_len == 2:
                            sides = zip(
                                [["top", "bottom"], ["right", "left"]],
                                setting_value
                            )

                            for side in sides:
                                for s in side[0]:
                                    self.borders[s].value = float(side[1])

                        if setting_len == 3:
                            self.borders["top"].value = float(setting_value[0])

                            for side in ["right", "left"]:
                                self.borders[side].value = float(
                                    setting_value[1]
                                )

                            self.borders["bottom"].value = float(
                                setting_value[2]
                            )

                        if setting_len == 4:
                            sides = zip(
                                ["top", "right", "bottom", "left"],
                                setting_value
                            )

                            for side in sides:
                                self.borders[side[0]].value = float(side[1])

                # -------------------------------------------------------------

    def set_orientation(self, orientation):
        """
        Set (master) page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: string
        :returns: boolean
        :rtype: boolean
        """

        if orientation.lower() in PageAbstract.orientation_xml.keys():
            self.orientation = orientation.lower()
            # FIXME TODO Modify (master) page box accordingly

            return True

        else:
            raise ValueError(
                "orientation parameter of set_orientation must be "\
                "portrait' or 'landscape'"
            )

    def fromxml(self, xml, master=False):
        """
        Set (master) page attributes according to LXML Element

        :param xml: (Master) page source as XML element
        :type xml: lxml._Element
        :param master: If the page is a master page or not
        :type master: boolean
        :returns: boolean
        :rtype: boolean
        """
        if master:
            tag = "MASTERPAGE"
        else:
            tag = "PAGE"

        if xml.tag != tag:
            return False

        #--- Box settings -------------------------------------

        posx = xml.get("PAGEXPOS")
        posy = xml.get("PAGEYPOS")
        dim_width = xml.get("PAGEWIDTH")
        dim_height = xml.get("PAGEHEIGHT")

        valid_box = 0

        for test in [posx, posy, dim_width, dim_height]:
            if test is not None:
                valid_box += 1

        if valid_box == 4:

            self.box.set_box(
                top_lx=posx, top_ly=posy,
                width=dim_width, height=dim_height
            )

        #--- Name / Master page name --------------------------

        self.name = xml.get("NAM")
        self.master_name = xml.get("MNAM")

        is_leftest = xml.get("LEFT")
        if is_leftest is not None:
            self.is_leftest = xmlc.num_to_bool(is_leftest)

        page_number = xml.get("NUM")

        if page_number is not None:
            # NOTE Page 0 is the first page so we adjust to a human
            # readable count
            self.number = int(page_number) + 1

        #--- Paper size name and orientation ------------------

        paper_size = xml.get("Size")
        if paper_size is not None:
            self.paper_size = paper_size

        orientation = xml.get("Orientation")
        if orientation is not None:

            for h,x in PageAbstract.orientation_xml.items():
                if orientation == x:
                    self.orientation = h
                    break

        #--- Borders ------------------------------------------

        for b in ["left", "top", "bottom", "right"]:
            att = "BORDER{}".format(b.upper())
            self.borders[b] = dimensions.Dim(float(xml.get(att)))

        #--- Guides -------------------------------------------

        for guide_type in ["vertical", "horizontal"]:

            att_name = "{}Guides".format(guide_type.capitalize())

            att = xml.get(att_name)
            if att is not None:
                if att:
                    # NOTE Example of att value : "42.5197 56.6929 "

                    guides_list = []
                    guides_values = [g.strip() for g in att.split(" ")]

                    for g in guides_values:

                        if g:

                            guides_list.append(
                                dimensions.Dim(float(g))
                            )

                    # self.guides[guide_type] = att
                    self.guides[guide_type] = guides_list

        #--- Auto guides --------------------------------------

        for case in [["lines", "horizontal"], ["columns", "vertical"]]:
            ag_count = xml.get("AG{}AutoCount".format(case[1]))
            ag_gap = xml.get("AG{}AutoGap".format(case[1]))
            ag_origin = xml.get("AG{}AutoRefer".format(case[1]))

            if ag_count is not None:
                ag_count = int(ag_count)
                self.auto_guides[case[0]]["count"] = ag_count

            if ag_gap is not None:
                ag_gap = float(ag_gap)
                self.auto_guides[case[0]]["gap"].value = ag_gap

            if ag_origin is not None:

                for human,code in PageAbstract.autoguides_origin_xml.items():
                    if ag_origin == code:
                        self.auto_guides[case[0]]["origin"] = human

        ag_selection = xml.get("AGSelection")
        if ag_selection is not None:
            agx, agy, agw, agh = [float(dim) for dim in ag_selection.split()]

            self.auto_guides["selection"]["posx"].value = agx
            self.auto_guides["selection"]["posy"].value = agy
            self.auto_guides["selection"]["width"].value = agw
            self.auto_guides["selection"]["height"].value = agh

        #--- PDF effects -------------------------------------------------

        duration = xml.get("pageEffectDuration")
        if duration is not None:
            self.effect["duration"].value = int(duration)

        view_duration = xml.get("pageViewDuration")
        if view_duration is not None:
            self.effect["view-duration"].value = int(view_duration)

        effect_type = xml.get("effectType")
        if effect_type is not None:
            for human, code in PageAbstract.effect_type_xml.items():
                if effect_type == code:
                    self.effect["type"] = human
                    break

        effect_lines = xml.get("Dm")
        if effect_lines is not None:
            for human, code in PageAbstract.effect_mobile_line_xml.items():
                if effect_lines == code:
                    self.effect["mobile-lines"] = human
                    break

        effect_source = xml.get("M")
        if effect_source is not None:
            for human, code in PageAbstract.effect_source_xml.items():
                if effect_source == code:
                    self.effect["source"] = human
                    break

        effect_direction = xml.get("Di")
        if effect_direction is not None:
            for human, code in PageAbstract.effect_direction_xml.items():
                if effect_direction == code:
                    self.effect["direction"] = human
                    break

        #--- FIXME This records undocumented attributes -------

        self.undocumented = xmlc.all_undocumented_to_python(xml)

        return True

    def toxml(self, master=False):
        """
        Returns (master) page as LXML Element

        :param master: If the page is a master page or not
        :type master: bool
        :returns: xml
        :rtype: lxml._Element
        """

        #--- Appropriate tag if master page or page -----------

        if master:
            tag = "MASTERPAGE"
        else:
            tag = "PAGE"

        xml = ET.Element(tag)

        # Position and dimensions -----------------------------

        xml.attrib["PAGEXPOS"] = self.box.coords["top-left"][0].toxmlstr()
        xml.attrib["PAGEYPOS"] = self.box.coords["top-left"][1].toxmlstr()

        xml.attrib["PAGEWIDTH"] = self.box.dims["width"].toxmlstr()
        xml.attrib["PAGEHEIGHT"] = self.box.dims["height"].toxmlstr()

        xml.attrib["LEFT"] = xmlc.bool_to_num(self.is_leftest)

        # Borders ---------------------------------------------

        for b in ["left", "right", "top", "bottom"]:
            att = "BORDER{}".format(b.upper())
            xml.attrib[att] = self.borders[b].toxmlstr(True)

         #--- Page number --------------------------------------

        if self.number > 0:
            # NOTE Page 0 is the first page so we adjust back
            # from human counting to computer counting
            xml.attrib["NUM"] = str(self.number - 1)

        #--- Page name and master page name -------------------

        xml.attrib["NAM"] = self.name
        xml.attrib["MNAM"] = self.master_name

        #--- (Master) page size name and orientation ----------

        xml.attrib["Size"] = self.paper_size
        xml.attrib["Orientation"] = PageAbstract.orientation_xml[self.orientation]

        #--- Guides -------------------------------------------

        for guide_type,guides in self.guides.items():
            att_name = "{}Guides".format(guide_type.capitalize())

            # NOTE Example of att value : "42.5197 56.6929 "

            guides_str = " ".join([g.toxmlstr(True) for g in guides])

            if guides_str.strip():
                xml.attrib[att_name] = guides_str
            else:
                xml.attrib[att_name] = ""

        #--- Auto guides --------------------------------------

        for case in [["lines", "horizontal"], ["columns", "vertical"]]:
            # Page lines / columns count
            # AGhorizontalAutoCount="2"
            # AGverticalAutoCount="3"
            count_att = "AG{}AutoCount".format(case[1])

            # Page lines / columns gap (gouttière)
            # AGhorizontalAutoGap="17.007874015748"
            # AGverticalAutoGap="22.6771653543307"
            gap_att = "AG{}AutoGap".format(case[1])

            # Page lines / columns origin
            # 0 = Page 1 = Margins
            # AGhorizontalAutoRefer="0"
            # AGverticalAutoRefer="0"
            orig_att = "AG{}AutoRefer".format(case[1])

            xml.attrib[count_att] = str(self.auto_guides[case[0]]["count"])
            xml.attrib[gap_att] = self.auto_guides[case[0]]["gap"].toxmlstr()
            xml.attrib[orig_att] = PageAbstract.autoguides_origin_xml[self.auto_guides[case[0]]["origin"]]

        # Check if auto guides if different from default value
        # If the X, Y, W, H of the selection are 0, then it is not custom

        custom_ag_selection = [
            dim.value
            for dim in self.auto_guides["selection"].values()
            if dim.value == 0.0
        ]
        custom_ag_selection = len(custom_ag_selection) != 4

        if custom_ag_selection:
            ag_selection = " ".join(
                [
                    mathc.truncate(self.auto_guides["selection"][number], 3)
                    for number in ["posx", "posy", "width", "height"]
                ]
            )

        else:
            ag_selection = "0 0 0 0"

        xml.attrib["AGSelection"] = ag_selection

        # --- PDF effects -------------------------------------------------

        xml.attrib["pageEffectDuration"] = self.effect["duration"].toxmlstr()
        xml.attrib["pageViewDuration"] = self.effect["view-duration"].toxmlstr()
        xml.attrib["effectType"] = PageAbstract.effect_type_xml[self.effect["type"]]
        xml.attrib["Dm"] = PageAbstract.effect_mobile_line_xml[self.effect["mobile-lines"]]
        xml.attrib["M"] = PageAbstract.effect_source_xml[self.effect["source"]]
        xml.attrib["Di"] = PageAbstract.effect_direction_xml[self.effect["direction"]]

        #--- FIXME This exports undocumented attributes -------

        try:
            xml, undoc_attribs = xmlc.all_undocumented_to_xml(
                xml, self.undocumented, True, tag, logger=logs.getLogger()
            )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromdefault(self, master=False, default=False):
        """
        Set (master) page attributes according to known defaults.

        :param master: If the page is a master page or not
        :param default: Name of the set of defaults ("a4", "letter")
        :type master: bool
        :type default: str
        """

        for border in self.borders.values():
            border.value = 40

        dim_default = "a4"

        if default:
            for format_name in ["a4", "letter"]:
                if default.lower() == format_name:
                    dim_default = format_name
                    break

        self.master_name = "Normal"

        self.box.dims["width"].fromdefault("{}-width".format(dim_default))
        self.box.dims["height"].fromdefault("{}-height".format(dim_default))

        self.box.set_box(
            top_lx=100.0, top_ly=20.0,
            width=self.box.dims["width"].value,
            height=self.box.dims["height"].value
        )


class Page(PageAbstract):
    """
    Page in SLA

    To add objects to this page, use Document.append()

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+---------------------------------+------------+
    | Kwargs       | Setting                         | Value type |
    +==============+=================================+============+
    | default      | Equivalent to a ``fromdefault`` | boolean    |
    |              | call.                           | or string  |
    |              |                                 |            |
    |              | Value being True or the name of |            |
    |              | the default set.                |            |
    +--------------+---------------------------------+------------+
    | posx         | X position of the page          | float      |
    +--------------+---------------------------------+------------+
    | posy         | Y position of the page          | float      |
    +--------------+---------------------------------+------------+
    | width        | Page width                      | float      |
    +--------------+---------------------------------+------------+
    | height       | Page height                     | float      |
    +--------------+---------------------------------+------------+
    | borders      | Shorthand for :                 | List of    |
    |              |                                 | floats     |
    |              | *rightborder*, *leftborder*,    |            |
    |              | *topborder*, *bottomborder*     |            |
    |              |                                 |            |
    |              | Read like the CSS margin        |            |
    |              | property:                       |            |
    |              |                                 |            |
    |              | **With 1 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & right & bottom & left]   |            |
    |              |                                 |            |
    |              | **With 2 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & bottom, right & left]    |            |
    |              |                                 |            |
    |              | **With 3 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right & left, bottom]     |            |
    |              |                                 |            |
    |              | **With 4 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right, bottom, left]      |            |
    +--------------+---------------------------------+------------+
    | rightborder  | Right border                    | float      |
    +--------------+---------------------------------+------------+
    | leftborder   | Left border                     | float      |
    +--------------+---------------------------------+------------+
    | topborder    | Top border                      | float      |
    +--------------+---------------------------------+------------+
    | bottomborder | Bottom border                   | float      |
    +--------------+---------------------------------+------------+
    """

    def __init__(self, **kwargs):
        PageAbstract.__init__(self)
        PageAbstract._quick_setup(self, kwargs)

    def toxml(self):
        """
        Returns page as LXML Element

        :returns: xml
        :rtype: lxml._Element
        """
        return PageAbstract.toxml(self, False)

    def fromxml(self, xml):
        """
        Set page attributes according to LXML Element

        :param xml: Page source as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.fromxml(self, xml, False)

    def fromdefault(self, default=False):
        """
        Set page attributes according to known defaults.

        :param default: Name of the set of defaults ("a4", "letter")
        :type default: str
        """
        PageAbstract.fromdefault(self, False, default)

    def set_orientation(self, orientation):
        """
        Set page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: str
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.set_orientation(self, orientation)


class MasterPage(PageAbstract):
    """
    Master page in SLA

    To add objects to this master page, use Document.append()

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+---------------------------------+------------+
    | Kwargs       | Setting                         | Value type |
    +==============+=================================+============+
    | default      | Equivalent to a ``fromdefault`` | boolean    |
    |              | call.                           | or string  |
    |              |                                 |            |
    |              | Value being True or the name of |            |
    |              | the default set.                |            |
    +--------------+---------------------------------+------------+
    | posx         | X position of the page          | float      |
    +--------------+---------------------------------+------------+
    | posy         | Y position of the page          | float      |
    +--------------+---------------------------------+------------+
    | width        | Page width                      | float      |
    +--------------+---------------------------------+------------+
    | height       | Page height                     | float      |
    +--------------+---------------------------------+------------+
    | borders      | Shorthand for :                 | List of    |
    |              |                                 | floats     |
    |              | *rightborder*, *leftborder*,    |            |
    |              | *topborder*, *bottomborder*     |            |
    |              |                                 |            |
    |              | Read like the CSS margin        |            |
    |              | property:                       |            |
    |              |                                 |            |
    |              | **With 1 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & right & bottom & left]   |            |
    |              |                                 |            |
    |              | **With 2 float in the list :**  |            |
    |              |                                 |            |
    |              | [top & bottom, right & left]    |            |
    |              |                                 |            |
    |              | **With 3 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right & left, bottom]     |            |
    |              |                                 |            |
    |              | **With 4 float in the list :**  |            |
    |              |                                 |            |
    |              | [top, right, bottom, left]      |            |
    +--------------+---------------------------------+------------+
    | rightborder  | Right border                    | float      |
    +--------------+---------------------------------+------------+
    | leftborder   | Left border                     | float      |
    +--------------+---------------------------------+------------+
    | topborder    | Top border                      | float      |
    +--------------+---------------------------------+------------+
    | bottomborder | Bottom border                   | float      |
    +--------------+---------------------------------+------------+
    """

    def __init__(self, **kwargs):
        PageAbstract.__init__(self)
        PageAbstract._quick_setup(self, kwargs)

    def toxml(self):
        """
        Returns master page object as LXML Element

        :returns: xml
        :rtype: lxml._Element
        """
        return PageAbstract.toxml(self, True)

    def fromxml(self, xml):
        """
        Set master page attributes according to LXML Element

        :param xml: Master page source as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.fromxml(self, xml, True)

    def fromdefault(self, default=False):
        """
        Set master page attributes according to known defaults.

        :param default: Name of the set of defaults ("a4", "letter")
        :type default: str
        """
        PageAbstract.fromdefault(self, True, default)

    def set_orientation(self, orientation):
        """
        Set master page orientation.

        :param orientation: "portrait" or "landscape"
        :type orientation: str
        :returns: bool
        :rtype: bool
        """
        return PageAbstract.set_orientation(self, orientation)


class PageSet(xmlc.PyScribusElement):
    """
    Page set object.

    DOCUMENT/PageSets/Set
    """

    DEFAULTS = {
        "Single Page": {"first": 0, "rows": 1, "columns": 1, "names": []},
        "Facing Pages": {
            "first": 1, "rows": 1, "columns": 2,
            "names": ["Left Page", "Right Page"]
        },
        "3-Fold": {
            "first": 0, "rows": 1, "columns": 3,
            "names": ["Left Page", "Middle", "Right Page"]
        },
        "4-Fold": {
            "first": 0, "rows": 1, "columns": 4,
            "names": [
                "Left Page", "Middle Left", "Middle Right", "Right Page"
            ]
        }
    }

    def __init__(self):
        xmlc.PyScribusElement.__init__(self)

        self.pyscribus_defaults = [k for k in PageSet.DEFAULTS.keys()]

        self.name = ""
        self.first_page = 0
        self.rows = 0
        self.columns = 0
        self.pages = []

    def fromdefault(self, default):
        """
        """

        if default in self.pyscribus_defaults:

            accurate = PageSet.DEFAULTS[default]

            self.name = default
            self.rows = accurate["rows"]
            self.pages = accurate["names"]
            self.columns = accurate["columns"]
            self.first_page = accurate["first"]

            return True
        else:
            return False

    def toxml(self):
        """
        :returns: xml
        :rtype: lxml._Element
        """

        xml = ET.Element("Set")

        #------------------------------------------------------

        xml.attrib["Name"] = self.name
        xml.attrib["FirstPage"] = str(self.first_page)
        xml.attrib["Rows"] = str(self.rows)
        xml.attrib["Columns"] = str(self.columns)

        #--- Page names ---------------------------------------

        if self.pages:

            for n in self.pages:

                pnx = ET.Element("PageNames")
                pnx.attrib["Name"] = n
                xml.append(pnx)

        #--- FIXME This exports undocumented attributes -------

        try:
            xml = xmlc.undocumented_to_xml(xml, self.undocumented, no_none=True)
        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromxml(self, xml):
        """
        :param xml: Page set as XML element
        :type xml: lxml._Element
        :returns: bool
        :rtype: bool
        """

        if xml.tag == "Set":
            return False

        #------------------------------------------------------

        name = xml.get("Name")
        if name is not None:
            self.name = name

        #------------------------------------------------------

        for att in ["FirstPage", "Rows", "Columns"]:

            atx = xml.get(att)
            if atx is not None:

                try:
                    atx = int(atx)
                except ValueError:
                    raise exceptions.InsaneSLAValue(
                        "Page set @{} must be a number.".format(
                            atx
                        )
                    )

        first = xml.get("FirstPage")
        if first is not None:
            self.first_page = int(first)

        rows = xml.get("Rows")
        if rows is not None:
            self.rows = int(rows)

        columns = xml.get("Columns")
        if columns is not None:
            self.columns = int(columns)

        #--- Page names ---------------------------------------

        for child in xml:
            if child.tag == "PageNames":
                n = child.get("Name")

                if n not in self.pages:
                    self.pages.append(n)

        #--- FIXME This records undocumented attributes -------

        self.undocumented = undocumented_to_python(
            xml, ["GapBelow", "GapHorizontal", "GapVertical"]
        )

        return True


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Functions for XML <-> Python data conversion.

Most of them are annoyingly simple, but necessary.
"""

# Imports ===============================================================#

import copy
import pprint

import lxml
import lxml.etree as ET

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

num_type_xml = {
    "decimal": "Type_1_2_3",
    "decimal-ar": "Type_1_2_3_ar",
    "lower-roman": "Type_i_ii_iii",
    "upper-roman": "Type_I_II_III",
    "lower-latin": "Type_a_b_c",
    "upper-latin": "Type_A_B_C",
    "alpha-ar": "Type_Alphabet_ar",
    "abjad-ar": "Type_Abjad_ar",
    "hebrew": "Type_Hebrew",
    "asterix": "Type_asterix",
    "cjk": "Type_CJK",
}

alignment = {
    "left": "0",
    "center": "1",
    "right": "2",
    "justify-left": "3",
    "justify": "4"
}

# Classes ===============================================================#

class PyScribusElement:
    """
    Abstract class for any element that must implement PyScribus
    common methods.

    :ivar list pyscribus_defaults: Names of default settings for
        fromdefault

    .. warning:: except listdefaults, all methods must be redefined by
        inherited classes.
    """

    def __init__(self):
        self.pyscribus_defaults = []

    #--- Quick setup ---------------------------------

    def _quick_setup(self, settings):
        """
        Method for defining style settings from class
        instanciation kwargs.

        In the inherited classes, the redefined method is called
        at class instanciation and PyScribusElement._quick_setup
        is called in the redefined method.

        Document settings keys in class docstring.

        Only treats "default" setting for inherited classes.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            # Call to fromdefault() to avoid conflicts with later settings

            if "default" in settings:

                if isinstance(settings["default"], bool):
                    self.fromdefault()
                else:
                    self.fromdefault(settings["default"])

    #--- Default(s) ----------------------------------

    def fromdefault(self):
        """
        Modify PyScribus element attributes according to a set of
        defaults.

        If fromdefault() has not other argument than self, there is
        only one, unnamed, set of default.

        If fromdefault() has a ``default`` argument, there is more
        than one set of default. Use listdefaults() to get the list
        of defaults available.

        :Example:

        .. code:: python

           # The class has only one default
           inherited_class_instance.fromdefault()

           # The class hass more than one default
           inherited_class_instance.fromdefault("english_default")

        """
        pass

    def listdefaults(self):
        """
        List available defaults for the current PyScribus object.

        .. note:: If this method returns an empty list, that
            doesn't mean there is no defaults : it could be mean
            there is **only one** default.

        .. seealso:: fromdefault()

        :rtype: list
        :returns: List of available defaults for fromdefault() method
        """

        return self.pyscribus_defaults

    #--- XML import / export -------------------------

    def toxml(self):
        """
        :rtype: lxml._Element
        :returns: SLA element as LXML ElementTree element
        """

        return False

    def fromxml(self, xml):
        """
        :type xml: lxml._Element
        :param xml: Some XML element
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        return False

    #--- Copy ----------------------------------------

    def copy(self, **kwargs):
        """
        Returns an independant copy of the instance.

        Use kwargs to quick set this copy as you made it.

        :type kwargs: dict
        :param kwargs: Quick setting (same as __init__)
        """

        duplicate = copy.deepcopy(self)

        if kwargs:
            duplicate._quick_setup(kwargs)

        return duplicate


class OrphanElement(PyScribusElement):
    """
    Abstract class for any element that must implement PyScribus
    common method and don't have any XML attribute & child.

    :type tag: str
    :param tag: XML tag

    :ivar string tag: XML tag
    """

    def __init__(self, tag):
        PyScribusElement.__init__(self)
        self.tag = tag

    def toxml(self):
        """
        :rtype: lxml._Element
        """
        xml = ET.Element(self.tag)
        return xml

    def fromdefault(self):
        # NOTE As this class implements XML elements without any attribute
        # or childs possible, an instanciation of OrphanElement is already
        # a default.
        return True

    def fromxml(self, xml):
        """
        Parse XML element with tag OrphanElement.tag

        :param xml: XML element
        :type xml: lxml._Element
        :rtype: bool
        :returns: bool
        """
        if xml.tag == self.tag:
            return True

        return False

    def __str__(self):
        return "<{}/>".format(self.tag)

# Fonctions =============================================================#

# We need a better typology of functions here :
#
# +---------------------+------------------+---------+---------------------------------+-----------+
# | Function name       | New name ?       | Returns | Manipulation                    | Side      |
# +---------------------+------------------+---------+---------------------------------+-----------+
# | float_or_int_string | out_cleanfloat   | string  | number without useless decimals | xml       |
# | str_or_nonestr      | out_str_nonestr  | string  | string or "None"                | xml ?     |
# | bool_to_num         | out_num_boolean  | string  | boolean as "0" or "1"           | xml       |
# | num_to_bool         | in_num_boolean   | boolean | "0" or "1" as boolean           | pyscribus |
# | bool_or_else_to_num | out_zero_string  | string  | "0" if not(v) or v              | xml       |
# | none_or_str         | out_empty_string | string  | "" if v is None or v            | xml ?     |
# +---------------------+------------------+---------+---------------------------------+-----------+

def float_or_int_string(f):

    if float(f) == int(f):
        return str(int(f))
    else:
        return str(f)

def str_or_nonestr(v):
    """
    Returns v if v is a non-empty string, else returns 'None'.

    :rtype: str
    """

    if v:
        return v
    else:
        return "None"

def bool_to_num(b):
    """
    Returns '1' if b is true, returns '0' if b is false.

    :rtype: str
    """

    if b:
        return "1"
    else:
        return "0"

def num_to_bool(n):
    """
    Return True if n == '1', else returns False.

    :rtype: bool
    """

    if n == "1":
        return True
    else:
        return False

def bool_or_else_to_num(b):
    """
    Returns '0' if b is False, else return b as string.

    :rtype: str
    """

    if b:
        return str(b)
    else:
        return "0"

def none_or_str(v):
    """
    If v is None, return empty string, else return v as string.

    :rtype: str
    """
    if v is None:
        return ""
    else:
        return str(v)

def undocumented_to_python(xml, attributes):
    """
    Function to manage the import of undocumented
    XML/SLA attributes.

    Used in PyScribus fromxml() methods until there is
    no attribute undocumented and handled.

    :param xml: XML element
    :type xml: lxml._Element
    :param attributes: List of attributes' names
    :type attributes: list
    :returns: Dictionnary with attributes names as
        keys and attribute values as values.
    :rtype: dict

    Use lxml._Element.get() to get attributes values,
    so missing attributes will have None as a value.

    .. seealso:: `undocumented_to_xml()`
    """
    undocumented = {}

    for att_name in attributes:
        undocumented[att_name] = xml.get(att_name)

    return undocumented

def undocumented_to_xml(xml, undocumented, no_none=False):
    """
    Function to manage the export of undocumented
    XML/SLA attributes.

    Used in PyScribus toxml() methods until there is
    no attribute undocumented and handled.

    :param xml: XML element
    :type xml: lxml._Element
    :param undocumented: Dictionnary of XML attributes.
        Return of undocumented_to_python()
    :type undocumented: dict
    :returns: xml
    :rtype: lxml._Element

    Add attributes to a XML element using undocumented
    dictionnary keys as XML attributes' names and
    undocumented dictionnary values as XML attributes'
    values.

    None values are ignored, through none_or_str().

    .. seealso:: undocumented_to_python()
    """

    for undatt, undvalue in undocumented.items():
        if no_none:
            if undvalue is None:
                continue

        xml.attrib[undatt] = none_or_str(undvalue)

    return xml

def all_undocumented_to_python(xml):
    """
    Function to manage the import of undocumented
    XML/SLA attributes.

    Used in PyScribus fromxml() methods until there is
    no attribute undocumented and handled.

    Used instead of undocumented_to_python() if there
    is no defined list of undocumented attributes.

    :param xml: XML element
    :type xml: lxml._Element
    :param attributes: List of attributes' names
    :type attributes: list
    :returns: Dictionnary with attributes names as
        keys and attribute values as values.
    :rtype: dict

    Get all xml attributes names and values so that
    all_undocumented_to_xml() will only have to filter
    those which aren't already in the LXML element to
    export.

    .. seealso:: all_undocumented_to_xml()
    """
    undocumented = {}

    for att_name, att_value in xml.attrib.items():
        undocumented[att_name] = att_value

    return undocumented

def all_undocumented_to_xml(
        xml, undocumented,
        report=True, msg="", passattr=[], logger=False):
    """
    Function to manage the export of undocumented
    XML/SLA attributes.

    Used in PyScribus toxml() methods until there is
    no attribute undocumented and handled.

    :param xml: XML element
    :type xml: lxml._Element
    :param undocumented: Dictionnary of XML attributes.
        Return of all_undocumented_to_python()
    :type undocumented: dict
    :type report: bool
    :param report: Print all undocumented attributes found in xml element.
    :type msg: str
    :param msg: Human readable name for the xml element when report parameter
        is True.
    :type passattr: list
    :param passattr: List of SLA attributes names to not report. For debug.
    :type logger: logging.Logger, bool
    :param logger: Logger object to use if report. If False, report is printed
        in STDOUT.
    :returns: List containing LXML element and undocumented attributes
        names list.
    :rtype: list

    Only adds undocumented attributes to xml element if
    xml element doesn't already have this attribute.

    .. seealso:: all_undocumented_to_python()
    """

    undoc_attribs = []

    for att_name, att_value in undocumented.items():

        if att_name not in xml.attrib:
            xml.attrib[att_name] = att_value

            if not passattr or att_name not in passattr:
                undoc_attribs.append(att_name)

    if report and undoc_attribs:

        msg = msg.strip()

        undocstr = "Undoc. XML attributes: {}: ".format(msg)

        if logger:
            logger.debug(
                undocstr + pprint.pformat(undoc_attribs, compact=True)
            )
        else:
            print(undocstr)
            pprint.pprint(undoc_attribs, width=80, compact=True)


    return [xml, undoc_attribs]

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

import setuptools

with open("../README.pypi.rst", "r") as fh:
    long_description = fh.read()

VERSION = "0.2.4"
DESCRIPTION = "Read, create and update Scribus .sla files. (Python < 3.8 version)"

REQUIRED = ['lxml', 'svg.path']

setuptools.setup(
    name="pyscribus-backported",
    version=VERSION,
    author="Étienne Nadji",
    author_email="etnadji@eml.cc",
    description=DESCRIPTION,
    long_description=long_description,
    packages=setuptools.find_packages(),
    platforms='any',
    license="GNU Affero General Public License v3 or later (AGPLv3+)",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Text Processing :: Markup :: XML",
        "Intended Audience :: Developers"
    ],
    project_urls={
        "Documentation": "https://etnadji.fr/pyscribus",
        "Source Code": "https://framagit.org/etnadji/pyscribus/-/tree/backported",
        "Issue tracker": "https://framagit.org/etnadji/pyscribus/-/issues",
    },
    python_requires='<=3.7.9',
    install_requires=REQUIRED,
    keywords=["scribus", "sla"],
)

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for styles manipulation.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from typing import Union, Optional, NoReturn, Literal
from enum import IntEnum
from pathlib import Path

import lxml
import lxml.etree as ET

import pyscribus.common.xml as xmlc
import pyscribus.common.math as mathc
import pyscribus.logs as logs
import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

StringOrPath = Union[str, Path]
BoolOrElement = Union[bool, ET.Element]

LeadingMode = Literal["fixed", "automatic", "grid"]

# Classes ===============================================================#

class TextDirection(IntEnum):
    """
    Enum for (paragraph) STYLE/@DIRECTION values.
    """

    LTR = 0
    RTL = 1


class NoteStyle(xmlc.PyScribusElement):
    """SLA notes style (notesStyle)"""

    range_type_xml = {
        "document": "0",
        "section": "1",
        "story": "2",
        "page": "3",
        "form": "4",
    }

    def __init__(self, default: bool=False):
        super().__init__()

        if default:
            self.fromdefault()
        else:
            self.name = ""
            self.set_type("footnote")
            self.num_start = 1
            self.num_type = "decimal"
            self.affixes = {"prefix": "", "suffix": ""}
            self.auto_dims = {"width": True, "height": True}
            self.auto_remove = True
            self.auto_link = True
            self.superscript = {"note": True, "mark": True}
            self.styles = {"mark": "", "note": ""}
            self.range = "document"

    def set_numtype(self, numtype: str) -> bool:
        # if numtype.lower() in NoteStyle.keys():
        if numtype.lower() in xmlc.num_type_xml.keys():
            self.num_type = numtype.lower()
            return True

        return False

    def set_type(self, note_type: str="footnote") -> NoReturn:
        if note_type.lower() == "footnote":
            self.is_endnotes = False
            self.is_footnotes = True
        else:
            self.is_endnotes = True
            self.is_footnotes = False

    def toxml(self) -> ET.Element:
        xml = ET.Element("notesStyle")

        xml.attrib["Name"] = self.name
        xml.attrib["Start"] = str(self.num_start)
        xml.attrib["Endnotes"] = xmlc.bool_to_num(self.is_endnotes)
        # xml.attrib["Type"] = NoteStyle.num_type_xml[self.num_type]
        xml.attrib["Type"] = xmlc.num_type_xml[self.num_type]
        xml.attrib["Range"] = NoteStyle.range_type_xml[self.range]
        xml.attrib["Prefix"] = self.affixes["prefix"]
        xml.attrib["Suffix"] = self.affixes["suffix"]
        xml.attrib["AutoHeight"] = xmlc.bool_to_num(self.auto_dims["height"])
        xml.attrib["AutoWidth"] = xmlc.bool_to_num(self.auto_dims["width"])
        xml.attrib["AutoRemove"] = xmlc.bool_to_num(self.auto_remove)
        xml.attrib["AutoWeld"] = xmlc.bool_to_num(self.auto_link)
        xml.attrib["SuperNote"] = xmlc.bool_to_num(self.superscript["note"])
        xml.attrib["SuperMaster"] = xmlc.bool_to_num(self.superscript["mark"])
        xml.attrib["MarksStyle"] = self.styles["mark"]
        xml.attrib["NotesStyle"] = self.styles["note"]

        return xml

    def fromxml(self, xml) -> bool:
        self.num_start = int(xml.get("Start"))
        self.name = xml.get("Name")
        self.affixes["prefix"] = xml.get("Prefix")
        self.affixes["suffix"] = xml.get("Suffix")
        self.auto_link = xmlc.num_to_bool(xml.get("AutoWeld"))
        self.auto_remove = xmlc.num_to_bool(xml.get("AutoRemove"))
        self.auto_dims["width"] = xml.get("AutoWidth")
        self.auto_dims["height"] = xml.get("AutoHeight")
        self.superscript["note"] = xmlc.num_to_bool(xml.get("SuperNote"))
        self.superscript["mark"] = xmlc.num_to_bool(xml.get("SuperMaster"))
        self.styles["note"] = xml.get("NotesStyle")
        self.styles["mark"] = xml.get("MarksStyle")

        nt = xml.get("Type")

        # for h,x in NoteStyle.num_type_xml.items():
        for h, x in xmlc.num_type_xml.items():
            if nt == x:
                self.num_type = h
                break

        rg = xml.get("Range")

        for h, x in NoteStyle.range_type_xml.items():
            if rg == x:
                self.range = h
                break

        return True

    def fromdefault(self) -> NoReturn:
        self.name = "Default"
        self.set_type("footnote")
        self.num_start = 1
        self.num_type = "decimal"
        self.affixes = {"prefix": "", "suffix": ")"}
        self.auto_dims = {"width": True, "height": True}
        self.auto_remove = True
        self.auto_link = True
        self.superscript = {"note": True, "mark": True}
        self.styles = {"mark": "", "note": ""}
        self.range = "document"

        return True


# FIXME This is a giant class.
# When styles attributes list will be stable, we will need to :
#   - move the type-specific (to|from)xml code and class attributes to
#     inherited classes. StyleAbstract.toxml() already returns xml Element.
#   - move the type-specific functions to inherited classes


class StyleAbstract(xmlc.PyScribusElement):
    """
    Abstract class for paragraph, character, table, cell styles in SLA.

    :type style_type: str
    :param style_type: Style type. Must be "paragraph", "character",
        "table", "cell".
    :type default: boolean
    :param default: Use default settings (False by default)
    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs tables of inherited classes)
    """

    # TODO Reste des attributs

    type_xml = {
        "paragraph": "STYLE",
        "character": "CHARSTYLE",
        "table": "TableStyle",
        "cell": "CellStyle",
    }

    name_xml = {"NAME": ["paragraph", "table", "cell"], "CNAME": ["character"]}

    shortcut_xml = {"PSHORTCUT": ["paragraph"], "SHORTCUT": ["character"]}

    parent_xml = {"PARENT": ["paragraph"], "CPARENT": ["character"]}

    fshade_xml = {
        "FillShade": ["table", "cell"],
        "FSHADE": ["character", "paragraph"],
    }

    fcolor_xml = {
        "FillColor": ["table", "cell"],
        "FCOLOR": ["character", "paragraph"],
    }

    default_name = {
        "paragraph": "Default Paragraph Style",
        "character": "Default Character Style",
        "table": "Default Table Style",
        "cell": "Default Cell Style",
    }

    bg_color_xml = {"BCOLOR": ["paragraph"], "BGCOLOR": ["character"]}
    bg_shade_xml = {"BSHADE": ["paragraph"], "BGSHADE": ["character"]}

    default_font = {
        "name": "Arial Regular",
        "size": 12,
        "features": [],
        "color": "Black",
        "alignment": "left",
    }

    def __init__(self, style_type, doc_parent, default=False, **kwargs):
        super().__init__()

        self.doc_parent = doc_parent

        if not style_type.lower() in StyleAbstract.type_xml.keys():
            raise ValueError("Unknown style type for StyleAbstract")

        self.style_type = style_type.lower()

        self.name = ""
        self.parent = None

        self.is_default = False

        if self.style_type in ["paragraph", "character"]:
            # TODO FIXME Handle font features as a dict instead
            # of a list, like in other styles
            # Features: -clig, inherit

            self.font = {
                "name": "",
                "size": 0,
                "features": [],
                "color": "",
                "alignment": "left",
                "strike": {"width": None, "offset": None},
                "underline": {"width": None, "offset": None},
                "shadow": {"hoffset": None, "voffset": None},
                # NOTE There is no offset setting for outline, obviously
                "outline": {"width": None},
            }

            # NOTE Weird, but FONTFEATURES is not the same as FEATURES
            # even if it has obviously an effect on fonts...
            self.features = {}

            self.lang = None
            self.shortcut = None

            self.background = {
                "color": "None",
                "shade": dimensions.Dim(100, unit="pc", integer=True)
            }

        if self.style_type in ["table", "cell", "character", "paragraph"]:
            self.fill = {"color": None, "shade": None}

        if self.style_type in ["table", "cell"]:
            self.borders = []

        if self.style_type == "cell":
            self.padding = {"left": 0, "right": 0, "top": 0, "bottom": 0}

        if self.style_type == "paragraph":
            self.tabs = []
            self.character_parent = ""

            self.space = {
                "after": None,
                "before": None,
            }

            self.leading = {"mode": None, "value": dimensions.Dim(0)}

        if default:
            self.fromdefault()

        if kwargs:
            self._quick_setup(kwargs)

    def _quick_setup(self, settings: dict) -> NoReturn:
        """
        Method for defining style settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if not settings:
            return

        xmlc.PyScribusElement._quick_setup(self, settings)

        for setting_name, setting_value in settings.items():

            if self.style_type in ["paragraph", "character"]:

                if setting_name == "defaultstyle":
                    if setting_value:
                        self.is_default = True
                    else:
                        self.is_default = False

            if setting_name == "name":
                self.name = setting_value

            if setting_name == "parent":
                self.parent = setting_value

            if self.style_type in ["paragraph", "character"]:

                if setting_name == "fontsize":
                    self.font["size"] = setting_value

                if setting_name == "font":
                    self.font["name"] = setting_value

    def _choose_att(self, atts_dict: dict) -> str:
        """
        Returns the correct XML attribute according to the
        StyleAbstract *class* attribute atts_dict.
        """
        for att, types in atts_dict.items():
            if self.style_type in types:
                return att

        return False

    def fromdefault(self) -> bool:
        self.name = StyleAbstract.default_name[self.style_type]

        if self.style_type in ["paragraph", "character"]:
            self.font = StyleAbstract.default_font

        # TODO Le reste

        return True

    def fromxml(self, xml: ET.Element, onlybool: bool=False) -> BoolOrElement:

        if xml.tag != StyleAbstract.type_xml[self.style_type]:
            return False

        # Name -------------------------------------------------

        if self.style_type in ["paragraph", "table", "cell", "character"]:
            name_att = self._choose_att(StyleAbstract.name_xml)

        name = xml.get(name_att)
        if name is not None:
            self.name = name

        # Common attributes ------------------------------------

        default = xml.get("DefaultStyle")
        if default is not None:
            self.is_default = xmlc.num_to_bool(default)

        # Specific attributes for paragraph --------------------
        # Moved in ParagraphStyle

        # Specific attributes for character --------------------
        # Moved in CharacterStyle

        # Specific attributes for table ------------------------

        if self.style_type == "table":
            pass

        # Specific attributes for cell -------------------------

        if self.style_type == "cell":

            for case in ["left", "right", "top", "bottom"]:
                att = xml.get("{}Padding".format(case.capitalize()))

                if att is not None:
                    self.padding[case] = att

        # Common attributes to paragraphs + characters ---------

        if self.style_type in ["paragraph", "character"]:

            bshade_att = self._choose_att(StyleAbstract.bg_shade_xml)
            bcolor_att = self._choose_att(StyleAbstract.bg_color_xml)

            back_color = xml.get(bcolor_att)
            if back_color is not None:
                self.background["color"] = back_color

            back_shade = xml.get(bshade_att)
            if back_shade is not None:
                self.background["shade"].value = float(back_shade)

            # Font and font features ---------------------------------

            font = xml.get("FONT")
            if font is not None:
                self.font["name"] = font

            font_size = xml.get("FONTSIZE")
            if font_size is not None:
                self.font["size"] = float(font_size)

            font_features = xml.get("FONTFEATURES")
            if font_features is not None:
                self.font["features"] = font_features.split(",")

            font_color = xml.get("FCOLOR")
            if font_color is not None:
                self.font["color"] = font_color

            # NOTE Weird FEATURES duplicate
            features = xml.get("FEATURES")
            if features is not None:
                self.features = features

            strike_width = xml.get("TXTSTW")
            if strike_width is not None:
                self.font["strike"]["width"] = dimensions.Dim(
                    float(strike_width), unit="pcdecim"
                )

            strike_offset = xml.get("TXTSTP")
            if strike_offset is not None:
                self.font["strike"]["offset"] = dimensions.Dim(
                    float(strike_offset), unit="pcdecim"
                )

            underline_width = xml.get("TXTULW")
            if underline_width is not None:
                self.font["underline"]["width"] = dimensions.Dim(
                    float(underline_width), unit="pcdecim"
                )

            underline_offset = xml.get("TXTULP")
            if underline_offset is not None:
                self.font["underline"]["offset"] = dimensions.Dim(
                    float(underline_offset), unit="pcdecim"
                )

            shadow_hoffset = xml.get("TXTSHX")
            if shadow_hoffset is not None:
                self.font["shadow"]["hoffset"] = dimensions.Dim(
                    float(shadow_hoffset), unit="pcdecim"
                )

            shadow_voffset = xml.get("TXTSHY")
            if shadow_voffset is not None:
                self.font["shadow"]["voffset"] = dimensions.Dim(
                    float(shadow_voffset), unit="pcdecim"
                )

            outline_width = xml.get("TXTOUT")
            if outline_width is not None:
                self.font["outline"]["width"] = dimensions.Dim(
                    float(outline_width), unit="pcdecim"
                )

            # Lang ---------------------------------------------------

            lang = xml.get("LANGUAGE")
            if lang is not None:
                if lang:
                    self.lang = lang

            # Parent style -------------------------------------------

            parent_att = self._choose_att(StyleAbstract.parent_xml)

            parent = xml.get(parent_att)
            if parent is not None:
                if parent:
                    self.parent = parent

            # Shortcut -----------------------------------------------

            shortcut_att = self._choose_att(StyleAbstract.shortcut_xml)

            shortcut = xml.get(shortcut_att)
            if shortcut = xml.get(shortcut_att)
                if shortcut:
                    self.shortcut = shortcut

        # Common attributes to tables, celles, characters ------

        if self.style_type in ["table", "cell", "character", "paragraph"]:
            fshade_att = self._choose_att(StyleAbstract.fshade_xml)
            fcolor_att = self._choose_att(StyleAbstract.fcolor_xml)

            fill_color = xml.get(fcolor_att)
            if fill_color is not None:
                self.fill["color"] = fill_color

            fill_shade = xml.get(fshade_att)
            if fill_shade is not None:
                self.fill["shade"] = float(fill_shade)

        # Childs -----------------------------------------------

        # Moved StyleTabs check in ParagraphStyle

        # ------------------------------------------------------

        # FIXME This records undocumented attributes -----------

        if self.style_type not in ["character", "paragraph"]:
            self.undocumented = xmlc.all_undocumented_to_python(xml)

        # ------------------------------------------------------

        # Returns lxml.etree._Element instead of True, for
        # inheritance purposes

        if onlybool:
            return True

        return xml

    def toxml(self) -> ET.Element:
        xml = ET.Element(StyleAbstract.type_xml[self.style_type])

        # Name -------------------------------------------------

        if self.style_type in ["paragraph", "table", "cell", "character"]:
            name_att = self._choose_att(StyleAbstract.name_xml)

        xml.attrib[name_att] = self.name

        # Common attributes ------------------------------------

        if self.is_default:
            xml.attrib["DefaultStyle"] = xmlc.bool_to_num(self.is_default)

        # Specific attributes for table ------------------------

        if self.style_type == "table":
            pass

        # Specific attributes for cell -------------------------

        if self.style_type == "cell":

            for case in ["left", "right", "top", "bottom"]:
                att_name = "{}Padding".format(case.capitalize())
                xml.attrib[att_name] = self.padding[case]

        # Common attributes to paragraphs + characters ---------

        if self.style_type in ["paragraph", "character"]:

            bshade_att = self._choose_att(StyleAbstract.bg_shade_xml)
            bcolor_att = self._choose_att(StyleAbstract.bg_color_xml)

            xml.attrib[bcolor_att] = self.background["color"]
            xml.attrib[bshade_att] = self.background["shade"].toxmlstr(True)

            if self.font["name"]:
                xml.attrib["FONT"] = self.font["name"]

            if self.style_type == "character":
                xml.attrib["FONTSIZE"] = str(self.font["size"])

            # TODO FIXME Handle font features as a dict instead
            # of a list, like in other styles
            # xml.attrib["FONTFEATURES"] = " ".join(self.font["features"].keys())

            if self.font["features"]:
                xml.attrib["FONTFEATURES"] = ",".join(self.font["features"])

            if self.font["color"]:
                xml.attrib["FCOLOR"] = self.font["color"]

            if self.font["strike"]["width"] is not None:
                xml.attrib["TXTSTW"] = self.font["strike"]["width"].toxmlstr(
                    True
                )

            if self.font["strike"]["offset"] is not None:
                xml.attrib["TXTSTP"] = self.font["strike"]["offset"].toxmlstr(
                    True
                )

            if self.font["shadow"]["hoffset"] is not None:
                xml.attrib["TXTSHX"] = self.font["shadow"]["hoffset"].toxmlstr(
                    True
                )

            if self.font["shadow"]["voffset"] is not None:
                xml.attrib["TXTSHY"] = self.font["shadow"]["voffset"].toxmlstr(
                    True
                )

            if self.font["outline"]["width"] is not None:
                xml.attrib["TXTOUT"] = self.font["outline"]["width"].toxmlstr(
                    True
                )

            if self.font["underline"]["width"] is not None:
                xml.attrib["TXTULW"] = self.font["underline"][
                    "width"
                ].toxmlstr(True)

            if self.font["underline"]["offset"] is not None:
                xml.attrib["TXTULP"] = self.font["underline"][
                    "offset"
                ].toxmlstr(True)

            if self.features:
                xml.attrib["FEATURES"] = self.features

            if self.lang is not None:
                xml.attrib["LANGUAGE"] = self.lang

            if self.shortcut is not None:
                shortcut_att = self._choose_att(StyleAbstract.shortcut_xml)
                xml.attrib[shortcut_att] = self.shortcut

            if self.parent is not None:
                # Parent style of same style type
                # CPARENT -> parent character style for this character style
                # PARENT  -> parent paragraph style for this paragraph style
                parent_att = self._choose_att(StyleAbstract.parent_xml)
                xml.attrib[parent_att] = self.parent

        # Common attributes to tables, celles, characters ------

        if self.style_type in ["table", "cell", "character"]:
            fshade_att = self._choose_att(StyleAbstract.fshade_xml)
            fcolor_att = self._choose_att(StyleAbstract.fcolor_xml)

            xml.attrib[fcolor_att] = xmlc.none_or_str(self.fill["color"])

            if self.fill["shade"] is not None:
                xml.attrib[fshade_att] = xmlc.float_or_int_string(
                    self.fill["shade"]
                )

        # ------------------------------------------------------

        # FIXME This exports undocumented attributes ----------

        try:
            if self.style_type not in ["paragraph", "character"]:
                xml, undoc_attribs = xmlc.all_undocumented_to_xml(
                    xml,
                    self.undocumented,
                    True,
                    self.style_type + " style " + self.name,
                    logger=logs.getLogger(),
                )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromstyle(self, style, override: bool=False):
        """
        Set attributes according to another style.

        :type style: ParagraphStyle, CharacterStyle, ...
        :param style: Style to load attributes from.
        :type override: boolean
        :param override: Override attributes that differs from style
        """
        # TODO
        pass


class ParagraphStyle(StyleAbstract):
    """
    Paragraph style in SLA (STYLE)

    :type default: boolean
    :param default: Use default settings (False by default)
    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+------------------------------------+
    | Kwargs       | Setting                            |
    +==============+====================================+
    | name         | Style name                         |
    +--------------+------------------------------------+
    | defaultstyle | Scribus default paragraph style?   |
    +--------------+------------------------------------+
    | parent       | Parent style name (paragraph)      |
    +--------------+------------------------------------+
    | cparent      | Parent style name (character)      |
    +--------------+------------------------------------+
    | alignment    | Text alignment. Must be in         |
    |              | ``pyscribus.common.xml.alignment`` |
    |              | keys                               |
    +--------------+------------------------------------+
    | direction    | Text direction. Either "ltr"       |
    |              | (left to right) or "rtl" (right to |
    |              | left.                              |
    +--------------+------------------------------------+
    | font         | Font name                          |
    +--------------+------------------------------------+
    | fontsize     | Font size                          |
    +--------------+------------------------------------+
    | spacebefore  | Space before paragraph             |
    +--------------+------------------------------------+
    | spaceafter   | Space after paragraph              |
    +--------------+------------------------------------+
    | leading      | Leading mode. Must be in           |
    |              | ``ParagraphStyle.leading_xml``.    |
    |              | If "fixed" mode, you must define   |
    |              | leadingValue setting.              |
    +--------------+------------------------------------+
    | leadingValue | Leading value in pica points if    |
    |              | leading mode is "fixed".           |
    +--------------+------------------------------------+
    | default      | Equivalent to a fromdefault        |
    |              | call, value being True or the      |
    |              | default name                       |
    +--------------+------------------------------------+

    :Example:

    .. code:: python

       title_style = styles.ParagraphStyle(
           name="Title", fontsize=18, alignment="center",
           leading="fixed", leadingValue=16
       )

       normal_style = styles.ParagraphStyle(
           name="Normal", fontsize=12, alignment="justify",
           leading="automatic"
       )

    """

    leading_xml = {
        "fixed": "0",
        "automatic": "1",
        "grid": "2",
    }

    glyph_scale_xml = {
        "max": "MaxGlyphExtend",
        "min": "MinGlyphShrink",
    }

    optical_margins_xml = {
        "both": 14,
        "left": 4,
        "right": 8,
        "none": 0,
        "undef": None
    }

    def __init__(self, doc_parent, default=False, **kwargs):
        StyleAbstract.__init__(self, "paragraph", doc_parent, default)

        self.indentations = {"left": None, "right": None, "first-line": None}

        self.listing = {"type": None, "char": None}

        self.glyph_scaling = {
            "min": dimensions.Dim(1.0, unit="pcscale"),
            "max": dimensions.Dim(1.0, unit="pcscale"),
        }

        self.initial = {
            "active": False,
            "lines": 2,
            "offset": dimensions.Dim(0),
            # NOTE The initial capital offset is documented as @DROPDIST but
            # implemented as @ParagraphEffectOffset so we store the attribute
            # used in the file for later.
            "__offset_att": "paragrapheffectoffset",
            # NOTE If there no auto_indent set, the value is not saved in SLA file.
            "auto_indent": False,
        }

        self.optical_margins = ParagraphStyle.optical_margins_xml["undef"]
        self.text_direction = TextDirection.LTR

        self._quick_setup(kwargs)

    def fromxml(self, xml: ET.Element) -> bool:
        is_paragraph = StyleAbstract.fromxml(self, xml, True)

        if not is_paragraph:
            return False

        # Text direction ------------------------------------------------

        text_direction = xml.get("DIRECTION")
        if text_direction is not None:

            if xmlc.num_to_bool(text_direction):
                self.text_direction = TextDirection.RTL
            else:
                self.text_direction = TextDirection.LTR

        # Optical margins -----------------------------------------------

        optic_margin = xml.get("OpticalMargins")

        for human, value in ParagraphStyle.optical_margins_xml.items():
            if value == optic_margin:
                self.optical_margins = human

        # Initial capital (aka drop capital) ----------------------------

        initial_used = xml.get("DROP")
        if initial_used is not None:
            self.initial["active"] = xmlc.num_to_bool(initial_used)

        initial_lines = xml.get("DROPLIN")
        if initial_lines is not None:
            self.initial["lines"] = int(initial_lines)

        initial_indent = xml.get("ParagraphEffectIndent")
        if initial_indent is not None:
            self.initial["auto_indent"] = xmlc.num_to_bool(initial_indent)

        for att_name in ["ParagraphEffectOffset", "DROPDIST"]:
            initial_offset = xml.get(att_name)
            if initial_offset is not None:
                self.initial["__offset_att"] = att_name
                self.initial["offset"].value = float(initial_offset)
                break

        # Glyph scaling -------------------------------------------------

        mgs = xml.get("MinGlyphShrink")
        if mgs is not None:
            self.glyph_scaling["min"].value = float(mgs)

        mge = xml.get("MaxGlyphExtend")
        if mge is not None:
            self.glyph_scaling["max"].value = float(mge)

        # Indentation ---------------------------------------------------

        # Left indentation of all paragraph lines

        left = xml.get("INDENT")
        if left is not None:
            self.indentations["left"] = dimensions.Dim(float(left))

        # Right indentation of all paragraph lines

        right = xml.get("RMARGIN")
        if right is not None:
            self.indentations["right"] = dimensions.Dim(float(right))

        # Indentation of the first line of the paragraph

        first = xml.get("FIRST")
        if first is not None:
            self.indentations["first-line"] = dimensions.Dim(
                float(first), negative=True
            )

        # Lists ---------------------------------------------------------

        is_ul, is_ol = False, False

        is_bullet = xml.get("Bullet")
        if is_bullet is not None:
            if is_bullet == "1":
                self.listing["type"] = "ul"
                is_ul = True

        is_numeroted = xml.get("Numeration")
        if is_numeroted is not None:
            if is_numeroted == "1":
                self.listing["type"] = "ol"
                is_ol = True

        if is_ol and is_ul:
            raise exceptions.InsaneSLAValue(
                "Style {} is both bullet and numeroted list.".format(self.name)
            )

        bullet_char = xml.get("BulletStr")
        if bullet_char is not None:
            self.listing["char"] = bullet_char

        # Parent character style ----------------------------------------

        character_parent = xml.get("CPARENT")
        if character_parent is not None:
            self.character_parent = character_parent

        # Leading -------------------------------------------------------

        leading = xml.get("LINESPMode")
        if leading is not None:

            for human, code in ParagraphStyle.leading_xml.items():
                if leading == code:
                    self.leading["mode"] = human
                    break

            if self.leading["mode"] == "fixed":
                leading_value = xml.get("LINESP")

                if leading_value is not None:
                    self.leading["value"].value = float(leading_value)

        # Alignment -----------------------------------------------------

        alignment = xml.get("ALIGN")
        if alignment is not None:
            for human, code in xmlc.alignment.items():
                if alignment == code:
                    self.font["alignment"] = human
                    break

        # Spaces before and after paragraph -----------------------------

        for case in [["VOR", "before"], ["NACH", "after"]]:

            space_att = xml.get(case[0])
            if space_att is not None:
                self.space[case[1]] = dimensions.Dim(float(space_att))

        # Tabs ----------------------------------------------------------

        for child in xml:

            if child.tag == "Tabs":
                to = StyleTab()

                success = to.fromxml(child)
                if success:
                    self.tabs.append(to)

        # End of parsing ------------------------------------------------

        self.undocumented = xmlc.all_undocumented_to_python(xml)

        return True

    def toxml(self) -> ET.Element:
        xml = StyleAbstract.toxml(self)

        # Text direction ------------------------------------------------

        xml.attrib["DIRECTION"] = str(int(self.text_direction))

        # Optical margins -----------------------------------------------

        if self.optical_margins is not None:
            xml.attrib["OpticalMargins"] = str(
                ParagraphStyle.optical_margins_xml[self.optical_margins]
            )

        # Initial capital (aka drop capital) ----------------------------

        xml.attrib["DROP"] = xmlc.bool_to_num(self.initial["active"])
        xml.attrib["DROPLIN"] = str(self.initial["lines"])

        xml.attrib[self.initial["__offset_att"]] = self.initial["offset"].toxmlstr()

        if self.initial["auto_indent"]:
            xml.attrib["ParagraphEffectIndent"] = xmlc.bool_to_num(self.initial["auto_indent"] )

        # Glyph scaling ---------------------------------------------

        for human, att in ParagraphStyle.glyph_scale_xml.items():
            if self.glyph_scaling[human].value != 1.0:
                xml.attrib[att] = self.glyph_scaling[human].toxmlstr()

        # -----------------------------------------------------------

        xml.attrib["ALIGN"] = xmlc.alignment[self.font["alignment"]]

        if self.leading["mode"] is not None:
            xml.attrib["LINESPMode"] = ParagraphStyle.leading_xml[
                self.leading["mode"]
            ]

            if self.leading["mode"] == "fixed":
                xml.attrib["LINESP"] = self.leading["value"].toxmlstr()

        # Indentation -----------------------------------------------

        for case in [
            ["left", "INDENT"],
            ["right", "RMARGIN"],
            ["first-line", "FIRST"],
        ]:

            if self.indentations[case[0]] is None:
                xml.attrib[case[1]] = "0"
            else:
                xml.attrib[case[1]] = self.indentations[case[0]].toxmlstr()

        # Lists ---------------------------------------------------------

        if self.listing["type"] == "ol":
            att_seq = [False, True]
        elif self.listing["type"] == "ul":
            att_seq = [True, False]
        else:
            att_seq = [False, False]

        for case in zip(att_seq, ["Bullet", "Numeration"]):
            xml.attrib[case[1]] = xmlc.bool_to_num(case[0])

        xml.attrib["BulletStr"] = xmlc.none_or_str(self.listing["char"])

        # Parent character style ------------------------------------

        # NOTE To not confuse with parent paragraph style which is
        # in @PARENT, and handled in StyleAbstract

        if self.character_parent:
            xml.attrib["CPARENT"] = self.character_parent

        for case in [["before", "VOR"], ["after", "NACH"]]:
            if self.space[case[0]] is not None:
                xml.attrib[case[1]] = self.space[case[0]].toxmlstr(True)

        # xml.attrib["VOR"] = self.space["before"].toxmlstr()
        # xml.attrib["NACH"] = self.space["after"].toxmlstr()

        # FONT SIZE -------------------------------------------------

        font_size = str(self.font["size"])

        if self.character_parent and font_size == "0":
            # If the font size of the paragraph is not defined
            # because it is defined in a parent character style,
            # don't write FONTSIZE attribute
            pass
        else:
            if font_size != "0":
                xml.attrib["FONTSIZE"] = font_size

        # -----------------------------------------------------------

        try:
            xml, undoc_attribs = xmlc.all_undocumented_to_xml(
                xml,
                self.undocumented,
                True,
                self.style_type + " style " + self.name,
                logger=logs.getLogger(),
            )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromdefault(self) -> bool:
        StyleAbstract.fromdefault(self)

        self.leading = {"mode": None, "value": dimensions.Dim(12)}

        return True

    def _quick_setup(self, settings: dict) -> NoReturn:
        """
        Method for defining style settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if not settings:
            return

        StyleAbstract._quick_setup(self, settings)

        for setting_name, setting_value in settings.items():

            if setting_name == "direction":
                setting_value = setting_value.strip().lower()

                if setting_value in ["ltr", "rtl"]:
                    if setting_value == "ltr":
                        self.text_direction = TextDirection.LTR
                    else:
                        self.text_direction = TextDirection.RTL
                else:
                    raise ValueError(
                        "Wrong direction setting '{}'. "
                        "Use either 'ltr' (left to right) 'rtl' "
                        "(right to left).".format(setting_value)
                    )

            if setting_name == "spacebefore":
                self.space["before"] = dimensions.Dim(float(setting_value))

            if setting_name == "spaceafter":
                self.space["after"] = dimensions.Dim(float(setting_value))

            if setting_name == "cparent":
                self.character_parent = setting_value

            if setting_name == "leading":
                if setting_value in ParagraphStyle.leading_xml:
                    self.leading["mode"] = setting_value

            if setting_name == "leadingValue":
                self.leading["value"] = dimensions.Dim(
                    float(setting_value)
                )

            if setting_name == "alignment":

                if setting_value.lower() in xmlc.alignment.keys():
                    self.font["alignment"] = setting_value
                else:
                    raise ValueError(
                        "Wrong alignement setting '{}'. "
                        "Alignement setting must be {}".format(
                            setting_value, ", ".join(xmlc.alignment.keys())
                        )
                    )

    def set_leadingmode(self, leadingmode: LeadingMode) -> bool:
        """
        Set leading mode.

        :type leadingmode: string
        :param leadingmode: "fixed", "automatic", "grid"
        """

        if leadingmode.lower() in ParagraphStyle.leading_xml:
            self.leading["mode"] = leadingmode
            return True

        raise ValueError()


class ParagraphStyle(StyleAbstract):
    """
    Paragraph style in SLA (STYLE)

    :type default: boolean
    :param default: Use default settings (False by default)
    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+------------------------------------+
    | Kwargs       | Setting                            |
    +==============+====================================+
    | name         | Style name                         |
    +--------------+------------------------------------+
    | defaultstyle | Scribus default paragraph style?   |
    +--------------+------------------------------------+
    | parent       | Parent style name (paragraph)      |
    +--------------+------------------------------------+
    | cparent      | Parent style name (character)      |
    +--------------+------------------------------------+
    | alignment    | Text alignment. Must be in         |
    |              | ``pyscribus.common.xml.alignment`` |
    |              | keys                               |
    +--------------+------------------------------------+
    | direction    | Text direction. Either "ltr"       |
    |              | (left to right) or "rtl" (right to |
    |              | left.                              |
    +--------------+------------------------------------+
    | font         | Font name                          |
    +--------------+------------------------------------+
    | fontsize     | Font size                          |
    +--------------+------------------------------------+
    | spacebefore  | Space before paragraph             |
    +--------------+------------------------------------+
    | spaceafter   | Space after paragraph              |
    +--------------+------------------------------------+
    | leading      | Leading mode. Must be in           |
    |              | ``ParagraphStyle.leading_xml``.    |
    |              | If "fixed" mode, you must define   |
    |              | leadingValue setting.              |
    +--------------+------------------------------------+
    | leadingValue | Leading value in pica points if    |
    |              | leading mode is "fixed".           |
    +--------------+------------------------------------+
    | default      | Equivalent to a fromdefault        |
    |              | call, value being True or the      |
    |              | default name                       |
    +--------------+------------------------------------+

    :Example:

    .. code:: python

       title_style = styles.ParagraphStyle(
           name="Title", fontsize=18, alignment="center",
           leading="fixed", leadingValue=16
       )

       normal_style = styles.ParagraphStyle(
           name="Normal", fontsize=12, alignment="justify",
           leading="automatic"
       )

    """

    leading_xml = {
        "fixed": "0",
        "automatic": "1",
        "grid": "2",
    }

    glyph_scale_xml = {
        "max": "MaxGlyphExtend",
        "min": "MinGlyphShrink",
    }

    optical_margins_xml = {
        "both": 14,
        "left": 4,
        "right": 8,
        "none": 0,
        "undef": None
    }

    def __init__(self, doc_parent, default=False, **kwargs):
        StyleAbstract.__init__(self, "paragraph", doc_parent, default)

        self.indentations = {"left": None, "right": None, "first-line": None}

        self.listing = {"type": None, "char": None}

        self.glyph_scaling = {
            "min": dimensions.Dim(1.0, unit="pcscale"),
            "max": dimensions.Dim(1.0, unit="pcscale"),
        }

        self.initial = {
            "active": False,
            "lines": 2,
            "offset": dimensions.Dim(0),
            # NOTE The initial capital offset is documented as @DROPDIST but
            # implemented as @ParagraphEffectOffset so we store the attribute
            # used in the file for later.
            "__offset_att": "paragrapheffectoffset",
            # NOTE If there no auto_indent set, the value is not saved in SLA file.
            "auto_indent": False,
        }

        self.optical_margins = ParagraphStyle.optical_margins_xml["undef"]
        self.text_direction = TextDirection.LTR

        self._quick_setup(kwargs)

    def fromxml(self, xml: ET.Element) -> bool:
        is_paragraph = StyleAbstract.fromxml(self, xml, True)

        if not is_paragraph:
            return False

        # Text direction ------------------------------------------------

        text_direction = xml.get("DIRECTION")
        if text_direction is not None:

            if xmlc.num_to_bool(text_direction):
                self.text_direction = TextDirection.RTL
            else:
                self.text_direction = TextDirection.LTR

        # Optical margins -----------------------------------------------

        optic_margin = xml.get("OpticalMargins")

        for human, value in ParagraphStyle.optical_margins_xml.items():
            if value == optic_margin:
                self.optical_margins = human

        # Initial capital (aka drop capital) ----------------------------

        initial_used = xml.get("DROP")
        if initial_used is not None:
            self.initial["active"] = xmlc.num_to_bool(initial_used)

        initial_lines = xml.get("DROPLIN")
        if initial_lines is not None:
            self.initial["lines"] = int(initial_lines)

        initial_indent = xml.get("ParagraphEffectIndent")
        if initial_indent is not None:
            self.initial["auto_indent"] = xmlc.num_to_bool(initial_indent)

        for att_name in ["ParagraphEffectOffset", "DROPDIST"]:
            initial_offset = xml.get(att_name)
            if initial_offset is not None:
                self.initial["__offset_att"] = att_name
                self.initial["offset"].value = float(initial_offset)
                break

        # Glyph scaling -------------------------------------------------

        mgs = xml.get("MinGlyphShrink")
        if mgs is not None:
            self.glyph_scaling["min"].value = float(mgs)

        mge = xml.get("MaxGlyphExtend")
        if mge is not None:
            self.glyph_scaling["max"].value = float(mge)

        # Indentation ---------------------------------------------------

        # Left indentation of all paragraph lines

        left = xml.get("INDENT")
        if left is not None:
            self.indentations["left"] = dimensions.Dim(float(left))

        # Right indentation of all paragraph lines

        right = xml.get("RMARGIN")
        if right is not None:
            self.indentations["right"] = dimensions.Dim(float(right))

        # Indentation of the first line of the paragraph

        first = xml.get("FIRST")
        if first is not None:
            self.indentations["first-line"] = dimensions.Dim(
                float(first), negative=True
            )

        # Lists ---------------------------------------------------------

        is_ul, is_ol = False, False

        is_bullet = xml.get("Bullet")
        if is_bullet is not None:
            if is_bullet == "1":
                self.listing["type"] = "ul"
                is_ul = True

        is_numeroted = xml.get("Numeration")
        if is_numeroted is not None:
            if is_numeroted == "1":
                self.listing["type"] = "ol"
                is_ol = True

        if is_ol and is_ul:
            raise exceptions.InsaneSLAValue(
                "Style {} is both bullet and numeroted list.".format(self.name)
            )

        bullet_char = xml.get("BulletStr")
        if bullet_char is not None:
            self.listing["char"] = bullet_char

        # Parent character style ----------------------------------------

        character_parent = xml.get("CPARENT")
        if character_parent is not None:
            self.character_parent = character_parent

        # Leading -------------------------------------------------------

        leading = xml.get("LINESPMode")
        if leading is not None:

            for human, code in ParagraphStyle.leading_xml.items():
                if leading == code:
                    self.leading["mode"] = human
                    break

            if self.leading["mode"] == "fixed":
                leading_value = xml.get("LINESP")

                if leading_value is not None:
                    self.leading["value"].value = float(leading_value)

        # Alignment -----------------------------------------------------

        alignment = xml.get("ALIGN")
        if alignment is not None:
            for human, code in xmlc.alignment.items():
                if alignment == code:
                    self.font["alignment"] = human
                    break

        # Spaces before and after paragraph -----------------------------

        for case in [["VOR", "before"], ["NACH", "after"]]:

            space_att = xml.get(case[0])
            if space_att is not None:
                self.space[case[1]] = dimensions.Dim(float(space_att))

        # Tabs ----------------------------------------------------------

        for child in xml:

            if child.tag == "Tabs":
                to = StyleTab()

                success = to.fromxml(child)
                if success:
                    self.tabs.append(to)

        # End of parsing ------------------------------------------------

        self.undocumented = xmlc.all_undocumented_to_python(xml)

        return True

    def toxml(self) -> ET.Element:
        xml = StyleAbstract.toxml(self)

        # Text direction ------------------------------------------------

        xml.attrib["DIRECTION"] = str(int(self.text_direction))

        # Optical margins -----------------------------------------------

        if self.optical_margins is not None:
            xml.attrib["OpticalMargins"] = str(
                ParagraphStyle.optical_margins_xml[self.optical_margins]
            )

        # Initial capital (aka drop capital) ----------------------------

        xml.attrib["DROP"] = xmlc.bool_to_num(self.initial["active"])
        xml.attrib["DROPLIN"] = str(self.initial["lines"])

        xml.attrib[self.initial["__offset_att"]] = self.initial["offset"].toxmlstr()

        if self.initial["auto_indent"]:
            xml.attrib["ParagraphEffectIndent"] = xmlc.bool_to_num(self.initial["auto_indent"] )

        # Glyph scaling ---------------------------------------------

        for human, att in ParagraphStyle.glyph_scale_xml.items():
            if self.glyph_scaling[human].value != 1.0:
                xml.attrib[att] = self.glyph_scaling[human].toxmlstr()

        # -----------------------------------------------------------

        xml.attrib["ALIGN"] = xmlc.alignment[self.font["alignment"]]

        if self.leading["mode"] is not None:
            xml.attrib["LINESPMode"] = ParagraphStyle.leading_xml[
                self.leading["mode"]
            ]

            if self.leading["mode"] == "fixed":
                xml.attrib["LINESP"] = self.leading["value"].toxmlstr()

        # Indentation -----------------------------------------------

        for case in [
            ["left", "INDENT"],
            ["right", "RMARGIN"],
            ["first-line", "FIRST"],
        ]:

            if self.indentations[case[0]] is None:
                xml.attrib[case[1]] = "0"
            else:
                xml.attrib[case[1]] = self.indentations[case[0]].toxmlstr()

        # Lists ---------------------------------------------------------

        if self.listing["type"] == "ol":
            att_seq = [False, True]
        elif self.listing["type"] == "ul":
            att_seq = [True, False]
        else:
            att_seq = [False, False]

        for case in zip(att_seq, ["Bullet", "Numeration"]):
            xml.attrib[case[1]] = xmlc.bool_to_num(case[0])

        xml.attrib["BulletStr"] = xmlc.none_or_str(self.listing["char"])

        # Parent character style ------------------------------------

        # NOTE To not confuse with parent paragraph style which is
        # in @PARENT, and handled in StyleAbstract

        if self.character_parent:
            xml.attrib["CPARENT"] = self.character_parent

        for case in [["before", "VOR"], ["after", "NACH"]]:
            if self.space[case[0]] is not None:
                xml.attrib[case[1]] = self.space[case[0]].toxmlstr(True)

        # xml.attrib["VOR"] = self.space["before"].toxmlstr()
        # xml.attrib["NACH"] = self.space["after"].toxmlstr()

        # FONT SIZE -------------------------------------------------

        font_size = str(self.font["size"])

        if self.character_parent and font_size == "0":
            # If the font size of the paragraph is not defined
            # because it is defined in a parent character style,
            # don't write FONTSIZE attribute
            pass
        else:
            if font_size != "0":
                xml.attrib["FONTSIZE"] = font_size

        # -----------------------------------------------------------

        try:
            xml, undoc_attribs = xmlc.all_undocumented_to_xml(
                xml,
                self.undocumented,
                True,
                self.style_type + " style " + self.name,
                logger=logs.getLogger(),
            )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml

    def fromdefault(self) -> bool:
        StyleAbstract.fromdefault(self)

        self.leading = {"mode": None, "value": dimensions.Dim(12)}

        return True

    def _quick_setup(self, settings: dict) -> NoReturn:
        """
        Method for defining style settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if not settings:
            return

        StyleAbstract._quick_setup(self, settings)

        for setting_name, setting_value in settings.items():

            if setting_name == "direction":
                setting_value = setting_value.strip().lower()

                if setting_value in ["ltr", "rtl"]:
                    if setting_value == "ltr":
                        self.text_direction = TextDirection.LTR
                    else:
                        self.text_direction = TextDirection.RTL
                else:
                    raise ValueError(
                        "Wrong direction setting '{}'. "
                        "Use either 'ltr' (left to right) 'rtl' "
                        "(right to left).".format(setting_value)
                    )

            if setting_name == "spacebefore":
                self.space["before"] = dimensions.Dim(float(setting_value))

            if setting_name == "spaceafter":
                self.space["after"] = dimensions.Dim(float(setting_value))

            if setting_name == "cparent":
                self.character_parent = setting_value

            if setting_name == "leading":
                if setting_value in ParagraphStyle.leading_xml:
                    self.leading["mode"] = setting_value

            if setting_name == "leadingValue":
                self.leading["value"] = dimensions.Dim(
                    float(setting_value)
                )

            if setting_name == "alignment":

                if setting_value.lower() in xmlc.alignment.keys():
                    self.font["alignment"] = setting_value
                else:
                    raise ValueError(
                        "Wrong alignement setting '{}'. "
                        "Alignement setting must be {}".format(
                            setting_value, ", ".join(xmlc.alignment.keys())
                        )
                    )

    def set_leadingmode(self, leadingmode: LeadingMode) -> bool:
        """
        Set leading mode.

        :type leadingmode: string
        :param leadingmode: "fixed", "automatic", "grid"
        """

        if leadingmode.lower() in ParagraphStyle.leading_xml:
            self.leading["mode"] = leadingmode
            return True

        raise ValueError()


class CharacterStyle(StyleAbstract):
    """
    Character style in SLA (CHARSTYLE)

    :type default: boolean
    :param default: Use default settings (False by default)
    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +--------------+-----------------------------+----------+
    | Kwargs       | Setting                     | Type     |
    +==============+=============================+==========+
    | name         | Style name                  | string   |
    +--------------+-----------------------------+----------+
    | defaultstyle | Scribus default character   | boolean  |
    |              | style ?                     | boolean  |
    +--------------+-----------------------------+----------+
    | parent       | Parent style name           | string   |
    +--------------+-----------------------------+----------+
    | font         | Font name                   | string   |
    +--------------+-----------------------------+----------+
    | fontsize     | Font size                   | string   |
    +--------------+-----------------------------+----------+
    | default      | Equivalent to a fromdefault | boolean, |
    |              | call, value being True or   | string   |
    |              | the default name.           |          |
    +--------------+-----------------------------+----------+
    """

    scale_xml = {
        "SCALEH": "horizontal",
        "SCALEV": "vertical",
        "BASEO": "baseline_offset"
    }

    def __init__(self, doc_parent, default=False, **kwargs):
        StyleAbstract.__init__(self, "character", doc_parent, default)
        StyleAbstract._quick_setup(self, kwargs)

        self.font["space_width"] = dimensions.Dim(1, unit="pcdecim")
        self.font["kerning"] = None

        self.hyphen_word_min = None

        self.scale = {
            "horizontal": 100,
            "vertical": 100,
            "baseline_offset": 0,
        }

        self.stroke = {
            "color": "Black",
            "shade": dimensions.Dim(100, unit="pc", integer=True),
        }

    def fromxml(self, xml: ET.Element) -> bool:
        is_character = StyleAbstract.fromxml(self, xml, True)

        if not is_character:
            return False

        # Stroke settings -----------------------------------------------

        stroke_color = xml.get("SCOLOR")
        if stroke_color is not None:
            self.stroke["color"] = stroke_color

        stroke_shade = xml.get("SSHADE")
        if stroke_shade is not None:
            self.stroke["shade"].value = float(stroke_shade)

        # Hyphenation ---------------------------------------------------

        hyphen_word = xml.get("HyphenWordMin")
        if hyphen_word is not None:
            try:
                word_length = int(hyphen_word)

                if word_length < 3:
                    raise exceptions.InsaneSLAValue(
                        "HyphenWordMin in '{}' character style must be an "
                        "integer >= 3.".format(self.name)
                    )

                self.hyphen_word_min = word_length

            except ValueError:
                raise exceptions.InsaneSLAValue(
                    "HyphenWordMin in '{}' character style must be an "
                    "integer >= 3.".format(self.name)
                )

        # Font settings -------------------------------------------------

        wordtrack = xml.get("wordTrack")
        if wordtrack is not None:
            self.font["space_width"] = dimensions.Dim(
                float(wordtrack), unit="pcdecim"
            )

        kerning = xml.get("KERN")
        if kerning is not None:
            self.font["kerning"] = dimensions.Dim(float(kerning), unit="pc")

        # Scales --------------------------------------------------------

        for att, human in CharacterStyle.scale_xml.items():
            scale_att = xml.get(att)
            if scale_att is not None:
                self.scale[human] = float(scale_att)

        # End of parsing ------------------------------------------------

        self.undocumented = xmlc.all_undocumented_to_python(xml)

        return True

    def toxml(self) -> ET.Element:
        xml = StyleAbstract.toxml(self)

        # Stroke settings -----------------------------------------------

        xml.attrib["SCOLOR"] = self.stroke["color"]
        xml.attrib["SSHADE"] = self.stroke["shade"].toxmlstr()

        # Hyphenation ---------------------------------------------------

        if self.hyphen_word_min is not None:
            xml.attrib["HyphenWordMin"] = str(self.hyphen_word_min)

        # Font settings -------------------------------------------------

        if self.font["space_width"] is not None:
            if self.font["space_width"].value != 1.0:
                xml.attrib["wordTrack"] = self.font["space_width"].toxmlstr()

        if self.font["kerning"] is not None:
            xml.attrib["KERN"] = self.font["kerning"].toxmlstr(True)

        # Scales --------------------------------------------------------

        for att, human in CharacterStyle.scale_xml.items():
            xml.attrib[att] = str(mathc.necessary_float(self.scale[human]))

        # ---------------------------------------------------------------

        try:
            xml, undoc_attribs = xmlc.all_undocumented_to_xml(
                xml,
                self.undocumented,
                True,
                self.style_type + " style " + self.name,
                logger=logs.getLogger(),
            )

        except AttributeError:
            # NOTE If fromxml was not used
            pass

        return xml


class TabularStyleAbstract(StyleAbstract):
    """Abstract middle class for Table and Cell styles."""

    def __init__(self, style_type, doc_parent, default=False, **kwargs):
        StyleAbstract.__init__(self, style_type, doc_parent, default)

    def fromxml(self, xml) -> bool:
        x = StyleAbstract.fromxml(self, xml)

        if len(x) == 0:
            return False

        for border in x:

            if border.tag in TableBorder.sides_xml.values():
                tb = TableBorder()

                success = tb.fromxml(border)
                if success:
                    self.borders.append(tb)

        return True

    def toxml(self) -> ET.Element:
        xml = StyleAbstract.toxml(self)

        for border in self.borders:
            bx = border.toxml()
            xml.append(bx)

        return xml


class TabularStyleAbstract(StyleAbstract):
    """Abstract middle class for Table and Cell styles."""

    def __init__(self, style_type, doc_parent, default=False, **kwargs):
        StyleAbstract.__init__(self, style_type, doc_parent, default)

    def fromxml(self, xml) -> bool:
        x = StyleAbstract.fromxml(self, xml)

        if len(x) == 0:
            return False

        for border in x:

            if border.tag in TableBorder.sides_xml.values():
                tb = TableBorder()

                success = tb.fromxml(border)
                if success:
                    self.borders.append(tb)

        return True

    def toxml(self) -> ET.Element:
        xml = StyleAbstract.toxml(self)

        for border in self.borders:
            bx = border.toxml()
            xml.append(bx)

        return xml


class TableBorder(xmlc.PyScribusElement):
    """
    Table border settings in SLA
    (TableBorderLeft, TableBorderRight, TableBorderTop, TableBorderBottom)

    :type side: str
    :param side: (Cell) Side of the border
    """

    sides_xml = {
        "left": "TableBorderLeft",
        "right": "TableBorderRight",
        "top": "TableBorderTop",
        "bottom": "TableBorderBottom",
    }

    def __init__(self, side: Optional[str]=False):
        super().__init__()

        if side:
            if side.lower() in TableBorder.sides_xml.keys():
                self.side = side.lower()
            else:
                self.side = False
        else:
            self.side = False

        self.lines = []

        self.width = None

    def fromxml(self, xml: ET.Element) -> bool:
        self.side = False

        # Find the side of the table border -------------------------

        if xml.tag in TableBorder.sides_xml.values():

            for h, x in TableBorder.sides_xml.items():
                if x == xml.tag:
                    self.side = h
                    break

        if not self.side:
            msg = "Invalid side for TableBorder"

            if self.style_type == "table":
                msg += " in TableStyle"

            if self.style_type == "cell":
                msg += " in CellStyle"

            raise exceptions.InsaneSLAValue(msg)

        # If the table border's side was found ----------------------

        # -------------------------------------------------

        width = xml.get("Width")
        if width is not None:
            self.width = dimensions.Dim(float(width))

        # -------------------------------------------------

        for line in xml:

            if line.tag == "TableBorderLine":
                lo = TableBorderLine()

                success = lo.fromxml(line)
                if success:
                    self.lines.append(lo)

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element(TableBorder.sides_xml[self.side])

        if self.width is not None:
            xml.attrib["Width"] = self.width.toxmlstr(True)

        for line in self.lines:
            lx = line.toxml()
            xml.append(lx)

        return xml


class TableBorderLine(xmlc.PyScribusElement):
    """
    Cell border line in SLA (TableBorderLine)
    """

    def __init__(self):
        super().__init__()

        self.width = 0
        self.pen_style = "1"
        self.color = "Black"
        self.shade = 100

    def fromdefault(self) -> NoReturn:
        self.width = 1
        self.pen_style = "1"
        self.color = "Black"
        self.shade = 100

    def fromxml(self, xml: ET.Element) -> bool:
        if xml.tag != "TableBorderLine":
            return False

        width = xml.get("Width")
        if width is not None:
            self.width = float(width)

        color = xml.get("Color")
        if color is not None:
            self.color = color

        shade = xml.get("Shade")
        if shade is not None:
            self.shade = int(shade)

        style = xml.get("PenStyle")
        if style is not None:
            self.pen_style = style

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element("TableBorderLine")

        if float(self.width) == int(self.width):
            xml_width = int(self.width)
        else:
            xml_width = self.width

        xml.attrib["Width"] = str(xml_width)
        xml.attrib["PenStyle"] = str(self.pen_style)
        xml.attrib["Color"] = self.color
        xml.attrib["Shade"] = str(self.shade)

        return xml


class TableStyle(TabularStyleAbstract):
    """
    Table style in  (TableStyle)
    """

    def __init__(self, doc_parent, default=False, **kwargs):
        TabularStyleAbstract.__init__(self, "table", doc_parent, default)


class CellStyle(TabularStyleAbstract):
    """
    Cell style in SLA (CellStyle)
    """

    def __init__(self, doc_parent, default=False, **kwargs):
        TabularStyleAbstract.__init__(self, "cell", doc_parent, default)


class StyleTab(xmlc.PyScribusElement):
    tab_type_xml = {
        "left": "0",
        "right": "1",
        "period": "2",
        "comma": "3",
        "center": "4",
    }

    def __init__(self):
        super().__init__()

        self.type = False
        self.fill = False
        self.position = dimensions.Dim(0, unit="pica")

    def fromxml(self, xml: ET.Element) -> bool:
        """"""

        tab_type = xml.get("Type")
        if tab_type is not None:
            try:
                for human, code in StyleTab.tab_type_xml.items():
                    if code == tab_type:
                        self.type = human
                        break

            except IndexError:
                self.type = "left"

        pos = xml.get("Pos")
        if pos is not None:
            pos = float(pos)
            self.position.value = pos

        fill = xml.get("Fill")
        if fill is not None:
            self.fill = fill

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element()
        xml.attrib["Type"] = StyleTab.tab_type_xml(self.type)
        xml.attrib["Pos"] = self.position.toxmlstr()
        xml.attrib["Fill"] = self.fill

        return xml

    def fromdefault(self) -> bool:
        self.type = "left"
        self.fill = False
        self.position = dimensions.Dim(0, unit="pica")

        return True


class RuleStyle(xmlc.PyScribusElement):
    """
    Rule style (filet) in SLA (MultiLine)
    """

    def __init__(self):
        super().__init__()

        self.name = None
        self.lines = []

    def fromxml(self, xml: ET.Element) -> bool:
        """"""

        if xml.tag != "MultiLine":
            return False

        name = xml.get("Name")
        if name is not None:
            self.name = name

        for element in xml:

            if element.tag == "SubLine":
                sl = RuleStyleLine()

                success = sl.fromxml(element)
                if success:
                    self.lines.append(sl)

        return True

    def toxml(self) -> ET.Element:
        """"""

        xml = ET.Element("MultiLine")
        xml.attrib["Name"] = self.name

        for line in self.lines:
            lx = line.toxml()
            xml.append(lx)

        return xml


class RuleStyleLine(xmlc.PyScribusElement):
    """
    Sub-line in rule style in SLA (MultiLine/SubLine)
    """

    # Line end human <-> XML
    lineend_xml = {"square": "16", "round": "32", "flat": "0"}

    # Line join human <-> XML
    linejoin_xml = {"pointy": "0", "bevelled": "64", "round": "128"}

    # Line style human <-> XML
    dash_xml = {
        # Most common and named line styles
        "line": "1",
        "dashed": "2",
        "dots": "3",
        "line-dash": "4",
        "line-dash-dash": "5",
        # Here be dragons, and a lot of untitled line styles
        "untitled-6": "6",
        "untitled-7": "7",
        "untitled-8": "8",
        "untitled-9": "9",
        "untitled-10": "10",
        "untitled-11": "11",
        "untitled-12": "12",
        "untitled-13": "13",
        "untitled-14": "14",
        "untitled-15": "15",
        "untitled-16": "16",
        "untitled-17": "17",
        "untitled-18": "18",
        "untitled-19": "19",
        "untitled-20": "20",
        "untitled-21": "21",
        "untitled-22": "22",
        "untitled-23": "23",
        "untitled-24": "24",
        "untitled-25": "25",
        "untitled-26": "26",
        "untitled-27": "27",
        "untitled-28": "28",
        "untitled-29": "29",
        "untitled-30": "30",
        "untitled-31": "31",
        "untitled-32": "32",
        "untitled-33": "33",
        "untitled-34": "34",
        "untitled-35": "35",
        "untitled-36": "36",
        "untitled-37": "37",
    }

    def __init__(self):
        super().__init__()

        self.color = None
        self.shortcut = ""

        self.width = None
        self.opacity = None

        self.end = None
        self.join = None
        self.style = None

    def fromdefault(self) -> NoReturn:
        self.end = "flat"
        self.join = "pointy"
        self.opacity = dimensions.Dim(100, unit="pc")

    def fromxml(self, xml: ET.Element) -> bool:
        """
        """

        if xml.tag != "SubLine":
            return False

        # --- Shortcut -----------------------------------------------

        shortcut = xml.get("Shortcut")
        if shortcut is not None:
            self.shortcut = shortcut

        # --- Color and opacity --------------------------------------

        color = xml.get("Color")
        if color is not None:
            self.color = color

        opacity = xml.get("Shade")
        if opacity is not None:
            self.opacity = dimensions.Dim(float(opacity), unit="pc")

        # --- Line width ---------------------------------------------

        width = xml.get("Width")
        if width is not None:
            self.width = dimensions.Dim(float(width))

        # --- Line style ---------------------------------------------

        dash = xml.get("Dash")
        if dash is not None:

            for human, code in RuleStyleLine.dash_xml.items():
                if code == dash:
                    self.style = human
                    break

        # --- Line end -----------------------------------------------

        line_end = xml.get("LineEnd")
        if line_end is not None:

            for human, code in RuleStyleLine.lineend_xml.items():
                if code == line_end:
                    self.end = human
                    break

        # --- Line join ----------------------------------------------

        line_join = xml.get("LineJoin")
        if line_join is not None:

            for human, code in RuleStyleLine.linejoin_xml.items():
                if code == line_join:
                    self.join = human
                    break

        return True

    def toxml(self) -> ET.Element:
        """
        """

        xml = ET.Element("SubLine")

        # --- Color and opacity --------------------------------------

        if self.color is not None:
            xml.attrib["Color"] = self.color

        if self.opacity is not None:
            xml.attrib["Shade"] = self.opacity.toxmlstr(True)
        else:
            xml.attrib["Shade"] = "100"

        # --- Line style, end, join ----------------------------------

        xml.attrib["Dash"] = RuleStyleLine.dash_xml[self.style]
        xml.attrib["LineEnd"] = RuleStyleLine.lineend_xml[self.end]
        xml.attrib["LineJoin"] = RuleStyleLine.linejoin_xml[self.join]

        # --- Line width ---------------------------------------------

        if self.width is not None:
            xml.attrib["Width"] = self.width.toxmlstr(True)

        # --- Shortcut -----------------------------------------------

        xml.attrib["Shortcut"] = self.shortcut

        return xml


# Fonctions =============================================================#


def fromSLA(filepath: StringOrPath="", slastring: str="") -> Union[dict, bool]:
    """
    Parse a SLA file / string and returns a dictionary of all styles in it.

    :type filepath: string
    :param filepath: SLA filepath
    :type slastring: string
    :param slastring: SLA XML string
    :rtype: dict,boolean
    :returns: Dictionary of styles usable by document.Document, or False
    """

    styles = {
        "note": [],
        "paragraph": [],
        "character": [],
        "table": [],
        "rule": [],
    }

    if filepath:
        xml = ET.parse(str(filepath)).getroot()
    else:
        xml = ET.fromstring(slastring)

    if len(xml):

        for style_type in [
            [TableStyle, "table", "DOCUMENT/TableStyle"],
            [ParagraphStyle, "paragraph", "DOCUMENT/STYLE"],
            [CharacterStyle, "character", "DOCUMENT/CHARSTYLE"],
            [NoteStyle, "note", "DOCUMENT/NotesStyles/notesStyle"],
            [RuleStyle, "rule", "DOCUMENT/MultiLine"],
        ]:

            tmp_styles = xml.findall(style_type[2])

            if len(tmp_styles):

                for style_item in tmp_styles:
                    si = style_type[0]()
                    success = si.fromxml(style_item)

                    if success:
                        styles[style_type[1]].append(si)

        return styles

    return False


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus module for PDF & printing settings / elements.
"""

# Imports ===============================================================#

import collections

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions

from pyscribus.common.xml import *

import pyscribus.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class PDFSettings(PyScribusElement):
    """
    Class for PDF export settings in SLA
    """

    imgcomp_method_xml = {
        "automatic": 0, "jpeg": 1, "flate": 2, "none": 3
    }

    imgcomp_quality_xml = {
        "max": 0, "high": 1, "medium": 2, "low": 3, "min": 4
    }

    def __init__(self):
        super().__init__()

        self.lpi = []

        self.text_compression = True
        self.image_compression = {
            "method": "automatic",
            "quality": "max"
        }

        #-------------------------------------------------------

        # Flag, set to 1 when Presentation effects should be used
        PresentMode="0"

        #-------------------------------------------------------

        # Flag, set to 1 when Downsample Images is checked in the PDF-Options Dialog
        RecalcPic="0"
        # Resolution for downsampling Images
        PicRes="300"
        # Resolution for embedded EPS-Pictures or PDF's
        Resolution="300"

        #-------------------------------------------------------

        # (optional) Flag, set to 1 when Output should be in RGB
        RGBMode="1"

        # TODO : fromxml / toxml of this dict

        self.profiles = {
            "colors": {
                # (optional) Flag, set to 1 when ICC-Profiles should be used
                # for solid colours
                # UseProfiles="0"
                "used": False,
                # (optional) ICC-Profile for solid colours
                # SolidP="sRGB IEC61966-2.1"
                "name": "sRGB IEC61966-2.1"
            },
            "images": {
                # (optional) Flag, set to 1 when ICC-Profiles should be used
                # for images
                # UseProfiles2="0"
                "used": False,
                # ICC-Profile for images ?
                # ImageP="sRGB IEC61966-2.1"
                "name": "sRGB IEC61966-2.1"
            },
            "printer": {
                # (optional) ICC-Profile for the Printer, only used for
                # PDF-X/3 output
                # PrintP="Fogra27L CMYK Coated Press"
                "name": "Fogra27L CMYK Coated Press"
            }
        }

        #-------------------------------------------------------

        self.bleeds = collections.OrderedDict()
        self.bleeds["top"] = dimensions.Dim(0),
        self.bleeds["left"] = dimensions.Dim(0),
        self.bleeds["right"] = dimensions.Dim(0),
        self.bleeds["bottom"] = dimensions.Dim(0),
        self.bleeds["document"] = False

        #-------------------------------------------------------

        bleedMarks="0"
        cropMarks="0"
        registrationMarks="0"
        colorMarks="0"
        docInfoMarks="0"
        markLength="20.0012598425197"
        markOffset="0"

        #-------------------------------------------------------

        self.encryption = {
            "pass": {"owner": "", "user": ""},
            "settings": {"permissions": "-4", "encrypted": False}
        }

        #-- FIXME Documented but to organize -------------------

        # (optional) Flag, set to 1 when the informations in the LPI tags should be used
        # for Linescreening
        UseLpi="0"
        # Binding for the PDF-Document 0 = Left Binding 1 = Right Binding
        Binding="0"
        # PDF-Version which should be generated 12 = PDF-X/3 13 = PDF-1.3 14 = PDF-1.4
        Version="14"
        # Flag, set to 1 when Generate Thumbnails is checked in the PDF-Options Dialog
        Thumbnails="0"
        # Flag, set to 1 when use PDF-Articles is checked in the PDF-Options Dialog
        Articles="0"
        # Flag, set to 1 when include Bookmarks is checked in the PDF-Options Dialog
        Bookmarks="0"

        #-- FIXME Undocumented, managed with undocumented funs -

        # ImagePr="0"
        # UseLayers="0"
        # UseSpotColors="1"
        # doMultiFile="0"
        # displayBookmarks="0"
        # displayFullscreen="0"
        # displayLayers="0"
        # displayThumbs="0"
        # hideMenuBar="0"
        # hideToolBar="0"
        # fitWindow="0"
        # openAfterExport="0"
        # PageLayout="0"
        # openAction=""
        # Intent="1"
        # Intent2="0"
        # InfoString=""
        # FontEmbedding="0"
        # Grayscale="0"
        # firstUse="1"
        # EmbedPDF="0"
        # MirrorH="0"
        # MirrorV="0"
        # Clip="0"
        # rangeSel="0"
        # rangeTxt=""
        # RotateDeg="0"

    def fromxml(self, xml):

        if xml.tag == "PDF":
            # TODO

            text_comp = xml.get("Compress")

            if text_comp is not None:
                self.text_compression = num_to_bool(text_comp)

            # Image compression : method and quality -----------

            for case in [
                    [xml.get("CMethod"), "method",
                    PDFSettings.imgcomp_method_xml],
                    [xml.get("Quality"), "quality",
                    PDFSettings.imgcomp_quality_xml]]:

                att = xml.get(case[0])
                if att is not None:
                    if int(att) in case[2].values():
                        for human,code in case[2].items():
                            self.image_compression[case[1]] = human
                            break
                    else:
                        raise exceptions.InsaneSLAValue(
                            "Unknown image compression {}.".format(case[1])
                        )

            # Bleed settings -----------------------------------

            for case in ["top", "left", "right", "bottom"]:
                att_name = "B{}".format(case.capitalize())

                att = xml.get(att_name)
                if att is not None:
                    self.bleeds[case][0].value = float(att)

            udb = xml.get("useDocBleeds")

            if udb is not None:
                self.bleeds["document"] = num_to_bool(udb)

            # PDF encryption settings---------------------------

            encrypt = xml.get("Encrypt")
            encrypt_perm = xml.get("Permissions")

            for case in [["owner", "PassOwner"], ["user", "PassUser"]]:
                attrib = xml.get(case[1])
                if attrib is not None:
                    self.encryption["pass"][case[0]] = attrib

            if encrypt is not None:
                self.encryption["settings"]["encrypted"] = num_to_bool(encrypt)

            if encrypt_perm is not None:
                self.encryption["settings"]["permissions"] = encrypt_perm

            # Line per inch settings ---------------------------

            for element in xml:
                if element.tag == "LPI":
                    lo = LPI()
                    success = lo.fromxml(element)

                    if success:
                        self.lpi.append(lo)

            #--- FIXME This records undocumented attributes -------

            self.undocumented = undocumented_to_python(
                xml,
                [
                    "ImagePr", "UseLayers", "UseSpotColors", "doMultiFile",
                    "displayBookmarks", "displayFullscreen", "displayLayers",
                    "displayThumbs", "hideMenuBar", "hideToolBar", "fitWindow",
                    "openAfterExport", "PageLayout", "openAction", "Intent",
                    "Intent2", "InfoString", "FontEmbedding", "Grayscale",
                    "firstUse", "EmbedPDF", "MirrorH", "MirrorV", "Clip",
                    "rangeSel", "rangeTxt", "RotateDeg",
                ]
            )
        else:
            return False

        return True

    def toxml(self):
        xml = ET.Element("PDF")

        xml.attrib["Compress"] = bool_to_num(self.text_compression)

        xml.attrib["CMethod"] = str(
            PDFSettings.imgcomp_method_xml[self.image_compression["method"]]
        )

        xml.attrib["Quality"] = str(
            PDFSettings.imgcomp_quality_xml[self.image_compression["quality"]]
        )

        # Bleed settings -----------------------------------

        for bleed_config,bleed_value in self.bleeds.items():

            if bleed_config != "document":
                att_name = "B{}".format(bleed_config.capitalize())
                # FIXME Sometimes raises errors (when PDF/@BTop="0") because
                # not a list
                try:
                    xml.attrib[att_name] = bleed_value[0].toxmlstr()
                except:
                    xml.attrib[att_name] = bleed_value.toxmlstr()

        xml.attrib["useDocBleeds"] = bool_to_num(self.bleeds["document"])

        # PDF encryption settings---------------------------

        xml.attrib["PassOwner"] = self.encryption["pass"]["owner"]
        xml.attrib["PassUser"] = self.encryption["pass"]["user"]
        xml.attrib["Permissions"] = self.encryption["settings"]["permissions"]
        xml.attrib["Encrypt"] = bool_to_num(
            self.encryption["settings"]["encrypted"]
        )

        # Line per Inch children ---------------------------

        for lo in self.lpi:
            lx = lo.toxml()

            if len(lx):
                xml.append(lx)

        #--- FIXME This exports undocumented attributes -------

        try:
            xml = undocumented_to_xml(xml, self.undocumented)
        except AttributeError:
            pass

        return xml

    def fromdefault(self):
        self.text_compression = True
        self.image_compression = {
            "method": "automatic",
            "quality": "max"
        }

        # Default line per each settings

        self.lpi = []

        for dlpi in LPI.DEFAULTS:
            l = LPI()
            l.fromdefault(dlpi[0])

            self.lpi.append(l)


class LPI(PyScribusElement):
    """
    Lines per Inch settings (LPI) in SLA
    """

    spot_xml = {
        "dot": "0",
        "line": "1",
        "round": "2",
        "ellipse": "3",
    }

    DEFAULTS = [
        ["Black", 133, 45, "ellipse"],
        ["Cyan", 133, 105, "ellipse"],
        ["Magenta", 133, 75, "ellipse"],
        ["Yellow", 133, 90, "ellipse"]
    ]

    def __init__(self):
        super().__init__()

        self.pyscribus_defaults = [l[0] for l in LPI.DEFAULTS]

        # Linescreening angle
        self.angle = dimensions.Dim(0, unit="deg")

        # Name of the Colour for which these settings are ment
        self.color = ""

        # How many lines per Inch are used
        self.frequency = dimensions.Dim(133, unit="lpi", integer=True)

        # Code for the used Spotfunction
        self.spot = "ellipse"

    #--- PyScribus standard methods ----------------------------

    def fromdefault(self, default):

        if default in self.pyscribus_defaults:

            self.color = default

            for lpd in LPI.DEFAULTS:

                if lpd[0] == default:
                    self.frequency.value = lpd[1]
                    self.angle.value = lpd[2]

                    if lpd[3] in LPI.spot_xml:
                        self.spot = lpd[3]
                    else:
                        self.spot = "ellipse"

                    break

    def fromxml(self, xml):

        if xml.tag == "LPI":

            color = xml.get("Color")
            if color is not None:
                self.color = color

            freq = xml.get("Frequency")
            if freq is not None:
                self.frequency.value = int(freq)

            angle = xml.get("Angle")
            if angle is not None:
                self.angle.value = float(angle)

            spot = xml.get("SpotFunction")
            if spot is not None:

                for human,code in LPI.spot_xml.items():
                    if code == spot:
                        self.spot = human
                        break

            return True
        else:
            return False

    def toxml(self):
        xml = ET.Element("LPI")

        xml.attrib["Color"] = self.color
        xml.attrib["Frequency"] = self.frequency.toxmlstr()
        xml.attrib["Angle"] = self.angle.toxmlstr()
        xml.attrib["SpotFunction"] = LPI.spot_xml[self.spot]

        return xml


class PrinterSettings(PyScribusElement):

    def __init__(self):
        super().__init__()

        self.mirror_pages = {
            "horizontal": False,
            "vertical": False
        }

        self.marks = {
            "crop": False,
            "bleed": False,
            "registration": False,
            "color": False
        }

        firstUse="1"

        # Print to file
        toFile="0"
        useAltPrintCommand="0"
        outputSeparations="0"

        # Whether to use spot colors
        useSpotColors="0"
        useColor="0"

        # Use ICC color profiles
        useICC="0"
        # Whether to have grey component replacement
        doGCR="0"

        doClip="0"
        setDevParam="0"

        includePDFMarks="0"

        # The postscrip level
        PSLevel="0"

        # Which printer description language
        PDLanguage="0"
        markLength="7.185302734375"
        markOffset="0"

        # --- Bleeds --------------------------------------------

        self.bleeds = collections.OrderedDict()
        self.bleeds["top"] = dimensions.Dim(0),
        self.bleeds["left"] = dimensions.Dim(0),
        self.bleeds["right"] = dimensions.Dim(0),
        self.bleeds["bottom"] = dimensions.Dim(0),
        self.bleeds["document"] = False

        # -------------------------------------------------------

        printer=""
        filename=""
        separationName=""
        printerCommand=""

    #--- PyScribus standard methods ----------------------------

    def fromxml(self, xml):

        if xml.tag == "Printer":

            # --- Page mirroring ------------------------------------

            mrh = xml.get("mirrorH")
            if mrh is not None:
                self.mirror_pages["horizontal"] = num_to_bool(mrh)

            mrv = xml.get("mirrorV")
            if mrv is not None:
                self.mirror_pages["vertical"] = num_to_bool(mrv)

            # --- Marks ---------------------------------------------

            for case in ["crop", "bleed", "registration", "color"]:
                att_name = "{}Marks".format(case)

                att = xml.get(att_name)
                if att is not None:
                    self.marks[case] = num_to_bool(att)

            # --- Printer bleeds ------------------------------------

            for case in ["top", "left", "right", "bottom"]:
                att_name = "Bleed{}".format(case.capitalize())

                att = xml.get(att_name)
                if att is not None:
                    self.bleeds[case][0].value = float(att)

            udb = xml.get("useDocBleeds")
            if udb is not None:
                self.bleeds["document"] = num_to_bool(udb)

            # -------------------------------------------------------

            # TODO

            return True
        else:
            return False

    def toxml(self):
        xml = ET.Element("Printer")

        # TODO

        for bleed_config,bleed_value in self.bleeds.items():

            if bleed_config != "document":
                att_name = "Bleed{}".format(bleed_config.capitalize())
                xml.attrib[att_name] = bleed_value[0].toxmlstr()

        xml.attrib["mirrorH"] = bool_to_num(self.mirror_pages["horizontal"])
        xml.attrib["mirrorV"] = bool_to_num(self.mirror_pages["vertical"])

        xml.attrib["useDocBleeds"] = bool_to_num(self.bleeds["document"])

        for k,v in self.marks.items():
            att_name = "{}Marks".format(k)

            xml.attrib[att_name] = bool_to_num(v)

        return xml

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
ISO 217, international system for paper size, additionnal to ISO 216.

Untrimmed paper sizes RA, SRA.

See iso216paper module for A, B formats.
See iso269paper module for C format (including DL).
"""

# Imports ===============================================================#

from pyscribus.common.math import FloatEnum, PICA_TO_MM

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Raw A -----------------------------------------

# 16RA0 3440 × 4880
# 8RA0 2440 × 3440
# 4RA0 1720 × 2440
# 2RA0 1220 × 1720
# RA0 860 × 1220
# RA1 610 × 860
# RA2 430 × 610
# RA3 305 × 430
# RA4 215 × 305

# Supplementary raw A ---------------------------

# 16SRA0 3600 × 5120
# 8SRA0 2560 × 3600
# 4SRA0 1800 × 2560
# 2SRA0 1280 × 1800
# SRA0 900 × 1280
# SRA1 640 × 900
# SRA2 450 × 640
# SRA3 320 × 450
# SRA4 225 × 320

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
ANSI normalized paper sizes.

ANSI A to E formats.

Yes, this is the module for Letter paper sizes, as Letter = ANSI A.
"""

# Imports ===============================================================#

from pyscribus.common.math import FloatEnum, PICA_TO_MM, INCH_TO_MM

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

# NOTE Dimensions are first converted into milimeters
#      because inches makes no sense to my european being.

class Letter(FloatEnum):
    """Letter format. Same as ANSI A."""

    WIDTH = (8.5 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (11 * INCH_TO_MM) / PICA_TO_MM

class A(FloatEnum):
    """ANSI A."""

    WIDTH = (8.5 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (11 * INCH_TO_MM) / PICA_TO_MM

class B(FloatEnum):
    """ANSI B."""

    WIDTH = (11 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (17 * INCH_TO_MM) / PICA_TO_MM

class C(FloatEnum):
    """ANSI C."""

    WIDTH = (17 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (22 * INCH_TO_MM) / PICA_TO_MM

class D(FloatEnum):
    """ANSI D."""

    WIDTH = (22 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (34 * INCH_TO_MM) / PICA_TO_MM

class E(FloatEnum):
    """ANSI E."""

    WIDTH = (34 * INCH_TO_MM) / PICA_TO_MM
    HEIGHT = (44 * INCH_TO_MM) / PICA_TO_MM

# vim:set shiftwidth=4 softtabstop=4 spl=en:
